from collections import defaultdict
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Optional, TypeVar, Union

import dagster._check as check
from dagster._core.definitions.dependency import (
    DependencyMapping,
    DependencyStructure,
    GraphNode,
    IDependencyDefinition,
    Node,
    NodeInvocation,
    OpNode,
)
from dagster._core.definitions.op_definition import OpDefinition
from dagster._core.errors import DagsterInvalidDefinitionError

if TYPE_CHECKING:
    from dagster._core.definitions.graph_definition import GraphDefinition
    from dagster._core.definitions.node_definition import NodeDefinition

T_DependencyKey = TypeVar("T_DependencyKey", str, "NodeInvocation")


def normalize_dependency_dict(
    dependencies: Optional[Union[DependencyMapping[str], DependencyMapping[NodeInvocation]]],
) -> DependencyMapping[NodeInvocation]:
    prelude = (
        'The expected type for "dependencies" is Union[Mapping[str, Mapping[str, '
        "DependencyDefinition]], Mapping[NodeInvocation, Mapping[str, DependencyDefinition]]]. "
    )

    if dependencies is None:
        return {}

    if not isinstance(dependencies, dict):
        raise DagsterInvalidDefinitionError(
            prelude + "Received value {dependencies} of type {type(dependencies)} at the top level."
        )

    normalized_dependencies: dict[NodeInvocation, Mapping[str, IDependencyDefinition]] = {}
    for key, dep_dict in dependencies.items():
        if not isinstance(dep_dict, dict):
            if isinstance(dep_dict, IDependencyDefinition):
                raise DagsterInvalidDefinitionError(
                    prelude
                    + f"Received a IDependencyDefinition one layer too high under key {key}. "
                    "The DependencyDefinition should be moved in to a dict keyed on "
                    "input name."
                )
            else:
                raise DagsterInvalidDefinitionError(
                    prelude
                    + f"Under key {key} received value {dep_dict} of type {type(dep_dict)}. "
                    "Expected dict[str, DependencyDefinition]"
                )

        for input_key, dep in dep_dict.items():
            if not isinstance(input_key, str):
                raise DagsterInvalidDefinitionError(
                    prelude + f"Received non-string key in the inner dict for key {key}. "
                    f"Unexpected inner dict key type: {type(input_key)}"
                )
            if not isinstance(dep, IDependencyDefinition):
                raise DagsterInvalidDefinitionError(
                    prelude
                    + f'Expected IDependencyDefinition for node "{key}" input "{input_key}". '
                    f"Received value {dep} of type {type(dep)}."
                )

        if isinstance(key, str):
            normalized_dependencies[NodeInvocation(key)] = dep_dict
        elif isinstance(key, NodeInvocation):
            normalized_dependencies[key] = dep_dict
        else:
            raise DagsterInvalidDefinitionError(
                prelude + "Expected str or NodeInvocation key in the top level dict. "
                "Received value {key} of type {type(key)}"
            )

    return normalized_dependencies


def create_execution_structure(
    node_defs: Sequence["NodeDefinition"],
    dependencies_dict: DependencyMapping[NodeInvocation],
    graph_definition: "GraphDefinition",
) -> tuple[DependencyStructure, Mapping[str, Node]]:
    """This builder takes the dependencies dictionary specified during creation of the
    JobDefinition object and builds (1) the execution structure and (2) a node dependency
    dictionary.

    For example, for the following dependencies:

    dep_dict = {
            NodeInvocation('giver'): {},
            NodeInvocation('sleeper', alias='sleeper_1'): {
                'units': DependencyDefinition('giver', 'out_1')
            },
            NodeInvocation('sleeper', alias='sleeper_2'): {
                'units': DependencyDefinition('giver', 'out_2')
            },
            NodeInvocation('sleeper', alias='sleeper_3'): {
                'units': DependencyDefinition('giver', 'out_3')
            },
            NodeInvocation('sleeper', alias='sleeper_4'): {
                'units': DependencyDefinition('giver', 'out_4')
            },
            NodeInvocation('total'): {
                'in_1': DependencyDefinition('sleeper_1', 'total'),
                'in_2': DependencyDefinition('sleeper_2', 'total'),
                'in_3': DependencyDefinition('sleeper_3', 'total'),
                'in_4': DependencyDefinition('sleeper_4', 'total'),
            },
        },

    This will create:

    node_dict = {
        'giver': <dagster._core.definitions.dependency.Node object>,
        'sleeper_1': <dagster._core.definitions.dependency.Node object>,
        'sleeper_2': <dagster._core.definitions.dependency.Node object>,
        'sleeper_3': <dagster._core.definitions.dependency.Node object>,
        'sleeper_4': <dagster._core.definitions.dependency.Node object>,
        'total': <dagster._core.definitions.dependency.Node object>
    }

    as well as a dagster._core.definitions.dependency.DependencyStructure object.
    """
    from dagster._core.definitions.graph_definition import GraphDefinition
    from dagster._core.definitions.node_definition import NodeDefinition

    check.sequence_param(node_defs, "node_defs", of_type=NodeDefinition)
    check.mapping_param(
        dependencies_dict,
        "dependencies_dict",
        key_type=(str, NodeInvocation),
        value_type=dict,
    )
    check.inst_param(graph_definition, "graph_definition", GraphDefinition)

    # Same as dep_dict but with NodeInvocation replaced by alias string
    aliased_dependencies_dict: dict[str, Mapping[str, IDependencyDefinition]] = {}

    # Keep track of node name -> all aliases used and alias -> name
    name_to_aliases: defaultdict[str, set[str]] = defaultdict(set)
    alias_to_node_invocation: dict[str, NodeInvocation] = {}
    alias_to_name: dict[str, str] = {}

    for node_invocation, input_dep_dict in dependencies_dict.items():
        # We allow deps of the form dependencies={'foo': DependencyDefinition('bar')}
        # Here, we replace 'foo' with NodeInvocation('foo')

        alias = node_invocation.alias or node_invocation.name

        name_to_aliases[node_invocation.name].add(alias)
        alias_to_node_invocation[alias] = node_invocation
        alias_to_name[alias] = node_invocation.name
        aliased_dependencies_dict[alias] = input_dep_dict

    node_dict = _build_graph_node_dict(
        node_defs, name_to_aliases, alias_to_node_invocation, graph_definition
    )

    _validate_dependencies(aliased_dependencies_dict, node_dict, alias_to_name)

    dependency_structure = DependencyStructure.from_definitions(
        node_dict, aliased_dependencies_dict
    )

    return dependency_structure, node_dict


def _build_graph_node_dict(
    node_defs: Sequence["NodeDefinition"],
    name_to_aliases: Mapping[str, set[str]],
    alias_to_node_invocation: Mapping[str, NodeInvocation],
    graph_definition,
) -> Mapping[str, Node]:
    from dagster._core.definitions.graph_definition import GraphDefinition

    nodes: list[Node] = []
    for node_def in node_defs:
        uses_of_node = name_to_aliases.get(node_def.name, {node_def.name})

        for alias in uses_of_node:
            node_invocation = alias_to_node_invocation.get(alias)

            node_invocation_tags = node_invocation.tags if node_invocation else {}
            hook_defs = node_invocation.hook_defs if node_invocation else frozenset()
            retry_policy = node_invocation.retry_policy if node_invocation else None
            node: Node
            if isinstance(node_def, GraphDefinition):
                node = GraphNode(
                    name=alias,
                    definition=node_def,
                    graph_definition=graph_definition,
                    tags=node_invocation_tags,
                    hook_defs=hook_defs,
                    retry_policy=retry_policy,
                )
            elif isinstance(node_def, OpDefinition):
                node = OpNode(
                    name=alias,
                    definition=node_def,
                    graph_definition=graph_definition,
                    tags=node_invocation_tags,
                    hook_defs=hook_defs,
                    retry_policy=retry_policy,
                )
            else:
                check.failed(f"Unexpected node_def type {node_def}")
            nodes.append(node)

    return {node.name: node for node in nodes}


def _validate_dependencies(
    dependencies: DependencyMapping[str],
    node_dict: Mapping[str, Node],
    alias_to_name: Mapping[str, str],
) -> None:
    for from_node, dep_by_input in dependencies.items():
        for from_input, dep_def in dep_by_input.items():
            for dep in dep_def.get_node_dependencies():
                if from_node == dep.node:
                    raise DagsterInvalidDefinitionError(
                        "Invalid dependencies: circular reference detected in node "
                        f'"{from_node}" input "{from_input}"'
                    )

                if from_node not in node_dict:
                    aliased_node = alias_to_name.get(from_node)
                    if aliased_node == from_node:
                        raise DagsterInvalidDefinitionError(
                            f'Invalid dependencies: node "{from_node}" in dependency dictionary not'
                            " found in node list"
                        )
                    else:
                        raise DagsterInvalidDefinitionError(
                            f'Invalid dependencies: node "{aliased_node}" (aliased by '
                            f'"{from_node}" in dependency dictionary) not found in node list'
                        )
                if not node_dict[from_node].definition.has_input(from_input):
                    from dagster._core.definitions.graph_definition import GraphDefinition

                    input_list = node_dict[from_node].definition.input_dict.keys()
                    node_type = (
                        "graph"
                        if isinstance(node_dict[from_node].definition, GraphDefinition)
                        else "op"
                    )
                    raise DagsterInvalidDefinitionError(
                        f'Invalid dependencies: {node_type} "{from_node}" does not have input '
                        f'"{from_input}". Available inputs: {list(input_list)}'
                    )

                if dep.node not in node_dict:
                    raise DagsterInvalidDefinitionError(
                        f'Invalid dependencies: node "{dep.node}" not found in node list. '
                        f'Listed as dependency for node "{from_node}" input "{from_input}" '
                    )

                if not node_dict[dep.node].definition.has_output(dep.output):
                    raise DagsterInvalidDefinitionError(
                        f'Invalid dependencies: node "{dep.node}" does not have output '
                        f'"{dep.output}". Listed as dependency for node "{from_node} input '
                        f'"{from_input}"'
                    )

                input_def = node_dict[from_node].definition.input_def_named(from_input)

                if dep_def.is_fan_in() and not input_def.dagster_type.supports_fan_in:
                    raise DagsterInvalidDefinitionError(
                        f'Invalid dependencies: for node "{dep.node}" input "{input_def.name}", the'
                        f' DagsterType "{input_def.dagster_type.display_name}" does not support'
                        " fanning in (MultiDependencyDefinition). Use the List type, since fanning"
                        " in will result in a list."
                    )



from collections.abc import Mapping
from functools import partial
from typing import Any, Optional, cast

import dagster._check as check
from dagster._annotations import beta_param
from dagster._core.definitions.asset_selection import AssetSelection, CoercibleToAssetSelection
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.run_request import SensorResult
from dagster._core.definitions.sensor_definition import (
    DefaultSensorStatus,
    SensorDefinition,
    SensorEvaluationContext,
    SensorType,
)
from dagster._core.definitions.utils import check_valid_name
from dagster._core.errors import DagsterInvalidInvocationError
from dagster._utils.tags import normalize_tags

MAX_ENTITIES = 500
EMIT_BACKFILLS_METADATA_KEY = "dagster/emit_backfills"
DEFAULT_AUTOMATION_CONDITION_SENSOR_NAME = "default_automation_condition_sensor"


def _evaluate(sensor_def: "AutomationConditionSensorDefinition", context: SensorEvaluationContext):
    from dagster._core.definitions.automation_tick_evaluation_context import (
        AutomationTickEvaluationContext,
    )
    from dagster._daemon.asset_daemon import (
        asset_daemon_cursor_from_instigator_serialized_cursor,
        asset_daemon_cursor_to_instigator_serialized_cursor,
    )

    asset_graph = check.not_none(context.repository_def).asset_graph
    cursor = asset_daemon_cursor_from_instigator_serialized_cursor(
        context.cursor,
        asset_graph,
    )

    evaluation_context = AutomationTickEvaluationContext(
        evaluation_id=cursor.evaluation_id,
        instance=context.instance,
        asset_graph=asset_graph,
        cursor=cursor,
        materialize_run_tags=sensor_def.run_tags,
        observe_run_tags={},
        auto_observe_asset_keys=set(),
        asset_selection=sensor_def.asset_selection,
        emit_backfills=sensor_def.emit_backfills,
        default_condition=sensor_def.default_condition,
        logger=context.log,
    )
    if evaluation_context.total_keys > MAX_ENTITIES:
        raise DagsterInvalidInvocationError(
            f'AutomationConditionSensorDefintion "{sensor_def.name}" targets {evaluation_context.total_keys} '
            f"assets or checks, which is more than the limit of {MAX_ENTITIES}. Either set `use_user_code_server` to `False`, "
            "or split this sensor into multiple AutomationConditionSensorDefinitions with AssetSelections that target fewer "
            "assets or checks."
        )

    run_requests, new_cursor, updated_evaluations = evaluation_context.evaluate()

    return SensorResult(
        run_requests=run_requests,
        cursor=asset_daemon_cursor_to_instigator_serialized_cursor(new_cursor),
        automation_condition_evaluations=updated_evaluations,
    )


def not_supported(context) -> None:
    raise NotImplementedError(
        "Automation condition sensors cannot be evaluated like regular user-space sensors."
    )


@beta_param(param="use_user_code_server")
@beta_param(param="default_condition")
class AutomationConditionSensorDefinition(SensorDefinition):
    """Targets a set of assets and repeatedly evaluates all the AutomationConditions on all of
    those assets to determine which to request runs for.

    Args:
        name: The name of the sensor.
        target (Union[str, Sequence[str], Sequence[AssetKey], Sequence[Union[AssetsDefinition, SourceAsset]], AssetSelection]):
            A selection of assets to evaluate AutomationConditions of and request runs for.
        tags (Optional[Mapping[str, str]]): A set of key-value tags that annotate the sensor and can
            be used for searching and filtering in the UI.
        run_tags (Optional[Mapping[str, Any]]): Tags that will be automatically attached to runs launched by this sensor.
        metadata (Optional[Mapping[str, object]]): A set of metadata entries that annotate the
            sensor. Values will be normalized to typed `MetadataValue` objects.
        default_status (DefaultSensorStatus): Whether the sensor starts as running or not. The default
            status can be overridden from the Dagster UI or via the GraphQL API.
        minimum_interval_seconds (Optional[int]): The frequency at which to try to evaluate the
            sensor. The actual interval will be longer if the sensor evaluation takes longer than
            the provided interval.
        description (Optional[str]): A human-readable description of the sensor.
        emit_backfills (bool): If set to True, will emit a backfill on any tick where more than one partition
            of any single asset is requested, rather than individual runs. Defaults to True.
        use_user_code_server (bool): (Beta) If set to True, this sensor will be evaluated in the user
            code server, rather than the AssetDaemon. This enables evaluating custom AutomationCondition
            subclasses, and ensures that the condition definitions will remain in sync with your user code
            version, eliminating version skew. Note: currently a maximum of 500 assets or checks may be
            targeted at a time by a sensor that has this value set.
        default_condition (Optional[AutomationCondition]): (Beta) If provided, this condition will
            be used for any selected assets or asset checks which do not have an automation condition defined.
            Requires `use_user_code_server` to be set to `True`.

    Examples:
        .. code-block:: python

            import dagster as dg

            # automation condition sensor that defaults to running
            defs1 = dg.Definitions(
                assets=...,
                sensors=[
                    dg.AutomationConditionSensorDefinition(
                        name="automation_condition_sensor",
                        target=dg.AssetSelection.all(),
                        default_status=dg.DefaultSensorStatus.RUNNING,
                    ),
                ]
            )

            # one automation condition sensor per group
            defs2 = dg.Definitions(
                assets=...,
                sensors=[
                    dg.AutomationConditionSensorDefinition(
                        name="raw_data_automation_condition_sensor",
                        target=dg.AssetSelection.groups("raw_data"),
                    ),
                    dg.AutomationConditionSensorDefinition(
                        name="ml_automation_condition_sensor",
                        target=dg.AssetSelection.groups("machine_learning"),
                    ),
                ]
            )

    """

    def __init__(
        self,
        name: str,
        *,
        target: CoercibleToAssetSelection,
        tags: Optional[Mapping[str, str]] = None,
        run_tags: Optional[Mapping[str, Any]] = None,
        default_status: DefaultSensorStatus = DefaultSensorStatus.STOPPED,
        minimum_interval_seconds: Optional[int] = None,
        description: Optional[str] = None,
        metadata: Optional[Mapping[str, object]] = None,
        emit_backfills: bool = True,
        use_user_code_server: bool = False,
        default_condition: Optional[AutomationCondition] = None,
    ):
        self._use_user_code_server = use_user_code_server
        check.bool_param(emit_backfills, "allow_backfills")

        self._default_condition = check.opt_inst_param(
            default_condition, "default_condition", AutomationCondition
        )
        check.param_invariant(
            not (self._default_condition and not self._use_user_code_server),
            "default_condition",
            "Setting a `default_condition` for a non-user-code AutomationConditionSensorDefinition is not supported.",
        )

        self._run_tags = normalize_tags(run_tags)

        # only store this value in the metadata if it's True
        if emit_backfills:
            metadata = {**(metadata or {}), EMIT_BACKFILLS_METADATA_KEY: True}

        super().__init__(
            name=check_valid_name(name),
            job_name=None,
            evaluation_fn=partial(_evaluate, self) if self._use_user_code_server else not_supported,
            minimum_interval_seconds=minimum_interval_seconds,
            description=description,
            job=None,
            jobs=None,
            default_status=default_status,
            required_resource_keys=None,
            asset_selection=target,
            tags=tags,
            metadata=metadata,
        )

    @property
    def run_tags(self) -> Mapping[str, str]:
        return self._run_tags

    @property
    def asset_selection(self) -> AssetSelection:
        return cast(AssetSelection, super().asset_selection)

    @property
    def emit_backfills(self) -> bool:
        return EMIT_BACKFILLS_METADATA_KEY in self.metadata

    @property
    def default_condition(self) -> Optional[AutomationCondition]:
        return self._default_condition

    @property
    def sensor_type(self) -> SensorType:
        return SensorType.AUTOMATION if self._use_user_code_server else SensorType.AUTO_MATERIALIZE



import dataclasses
import json
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, NamedTuple, Optional

from dagster._core.definitions.asset_key import EntityKey, T_EntityKey
from dagster._core.definitions.base_asset_graph import BaseAssetGraph
from dagster._core.definitions.events import AssetKey
from dagster._serdes.serdes import (
    FieldSerializer,
    JsonSerializableValue,
    PackableValue,
    SerializableNonScalarKeyMapping,
    UnpackContext,
    WhitelistMap,
    pack_value,
    unpack_value,
    whitelist_for_serdes,
)

if TYPE_CHECKING:
    from dagster._core.definitions.declarative_automation.serialized_objects import (
        AutomationConditionCursor,
        AutomationConditionEvaluationState,
        AutomationConditionNodeSnapshot,
    )


@whitelist_for_serdes(storage_name="AssetConditionCursorExtras")
class AutomationConditionCursorExtras(NamedTuple):
    """Represents additional state that may be optionally saved by an AutomationCondition between
    evaluations.
    """

    condition_snapshot: "AutomationConditionNodeSnapshot"
    extras: Mapping[str, PackableValue]


class ObserveRequestTimestampSerializer(FieldSerializer):
    def pack(
        self,
        mapping: Mapping[str, float],
        whitelist_map: WhitelistMap,
        descent_path: str,
    ) -> JsonSerializableValue:
        return pack_value(SerializableNonScalarKeyMapping(mapping), whitelist_map, descent_path)

    def unpack(
        self,
        unpacked_value: JsonSerializableValue,
        whitelist_map: WhitelistMap,
        context: UnpackContext,
    ) -> PackableValue:
        return unpack_value(unpacked_value, dict, whitelist_map, context)


@whitelist_for_serdes(
    field_serializers={
        "last_observe_request_timestamp_by_asset_key": ObserveRequestTimestampSerializer
    }
)
@dataclass(frozen=True)
# TODO: rename to scheduling cursor or something
# 2024-05-16 -- schrockn
class AssetDaemonCursor:
    """State that's stored between daemon evaluations.

    Args:
        evaluation_id (int): The ID of the evaluation that produced this cursor.
        previous_evaluation_state (Sequence[AutomationConditionEvaluationState]): (DEPRECATED) The
            evaluation info recorded for each asset on the previous tick.
        previous_cursors (Sequence[AutomationConditionCursor]): The cursor objects for each asset
            recorded on the previous tick.
    """

    evaluation_id: int
    last_observe_request_timestamp_by_asset_key: Mapping[AssetKey, float]

    previous_evaluation_state: Optional[Sequence["AutomationConditionEvaluationState"]]
    previous_condition_cursors: Optional[Sequence["AutomationConditionCursor"]] = None

    @staticmethod
    def empty(evaluation_id: int = 0) -> "AssetDaemonCursor":
        return AssetDaemonCursor(
            evaluation_id=evaluation_id,
            previous_evaluation_state=None,
            previous_condition_cursors=[],
            last_observe_request_timestamp_by_asset_key={},
        )

    @cached_property
    def previous_condition_cursors_by_key(
        self,
    ) -> Mapping[EntityKey, "AutomationConditionCursor"]:
        """Efficient lookup of previous cursor by asset key."""
        from dagster._core.definitions.declarative_automation.serialized_objects import (
            AutomationConditionCursor,
        )

        if self.previous_condition_cursors is None:
            # automatically convert AutomationConditionEvaluationState objects to AutomationConditionCursor
            return {
                evaluation_state.asset_key: AutomationConditionCursor.backcompat_from_evaluation_state(
                    evaluation_state
                )
                for evaluation_state in self.previous_evaluation_state or []
            }
        else:
            return {cursor.key: cursor for cursor in self.previous_condition_cursors}

    def get_previous_condition_cursor(
        self, key: T_EntityKey
    ) -> Optional["AutomationConditionCursor[T_EntityKey]"]:
        """Returns the AutomationConditionCursor associated with the given asset key. If no stored
        cursor exists, returns an empty cursor.
        """
        return self.previous_condition_cursors_by_key.get(key)

    def with_updates(
        self,
        evaluation_id: int,
        evaluation_timestamp: float,
        newly_observe_requested_asset_keys: Sequence[AssetKey],
        condition_cursors: Sequence["AutomationConditionCursor"],
    ) -> "AssetDaemonCursor":
        # do not "forget" about values for non-evaluated assets
        new_condition_cursors = dict(self.previous_condition_cursors_by_key)
        for cursor in condition_cursors:
            new_condition_cursors[cursor.key] = cursor

        return dataclasses.replace(
            self,
            evaluation_id=evaluation_id,
            previous_evaluation_state=[],
            previous_condition_cursors=list(new_condition_cursors.values()),
            last_observe_request_timestamp_by_asset_key={
                **self.last_observe_request_timestamp_by_asset_key,
                **{
                    asset_key: evaluation_timestamp
                    for asset_key in newly_observe_requested_asset_keys
                },
            },
        )

    def __hash__(self) -> int:
        return hash(id(self))


# BACKCOMPAT


def backcompat_deserialize_asset_daemon_cursor_str(
    cursor_str: str, asset_graph: Optional[BaseAssetGraph], default_evaluation_id: int
) -> AssetDaemonCursor:
    """This serves as a backcompat layer for deserializing the old cursor format. Will only recover
    the previous evaluation id.
    """
    data = json.loads(cursor_str)

    if isinstance(data, list):
        evaluation_id = data[0] if isinstance(data[0], int) else default_evaluation_id
        return AssetDaemonCursor.empty(evaluation_id)
    elif not isinstance(data, dict):
        return AssetDaemonCursor.empty(default_evaluation_id)
    elif asset_graph is None:
        return AssetDaemonCursor.empty(data.get("evaluation_id", default_evaluation_id))

    return AssetDaemonCursor(
        evaluation_id=data.get("evaluation_id") or default_evaluation_id,
        previous_evaluation_state=[],
        previous_condition_cursors=[],
        last_observe_request_timestamp_by_asset_key={},
    )


@whitelist_for_serdes
class LegacyAssetDaemonCursorWrapper(NamedTuple):
    """Wrapper class for the legacy AssetDaemonCursor object, which is not a serializable NamedTuple."""

    serialized_cursor: str

    def get_asset_daemon_cursor(self, asset_graph: Optional[BaseAssetGraph]) -> AssetDaemonCursor:
        return backcompat_deserialize_asset_daemon_cursor_str(
            self.serialized_cursor, asset_graph, 0
        )



import os
from collections.abc import Mapping, Sequence
from datetime import datetime
from typing import Any, Dict, Generic, List, NamedTuple, Optional, Union, cast  # noqa: F401, UP035

from typing_extensions import TypeAlias, TypeVar

import dagster._check as check
from dagster._annotations import PublicAttr, deprecated, deprecated_param
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.metadata.metadata_set import (
    NamespacedMetadataSet as NamespacedMetadataSet,
    TableMetadataSet as TableMetadataSet,
)
from dagster._core.definitions.metadata.metadata_value import (
    BoolMetadataValue as BoolMetadataValue,
    DagsterAssetMetadataValue as DagsterAssetMetadataValue,
    DagsterJobMetadataValue as DagsterJobMetadataValue,
    DagsterRunMetadataValue as DagsterRunMetadataValue,
    FloatMetadataValue as FloatMetadataValue,
    IntMetadataValue as IntMetadataValue,
    JsonMetadataValue as JsonMetadataValue,
    MarkdownMetadataValue as MarkdownMetadataValue,
    MetadataValue as MetadataValue,
    NotebookMetadataValue as NotebookMetadataValue,
    NullMetadataValue as NullMetadataValue,
    PathMetadataValue as PathMetadataValue,
    PoolMetadataValue as PoolMetadataValue,
    PythonArtifactMetadataValue as PythonArtifactMetadataValue,
    TableColumnLineageMetadataValue as TableColumnLineageMetadataValue,
    TableMetadataValue as TableMetadataValue,
    TableSchemaMetadataValue as TableSchemaMetadataValue,
    TextMetadataValue as TextMetadataValue,
    TimestampMetadataValue as TimestampMetadataValue,
    UrlMetadataValue as UrlMetadataValue,
)
from dagster._core.definitions.metadata.source_code import (
    DEFAULT_SOURCE_FILE_KEY as DEFAULT_SOURCE_FILE_KEY,
    AnchorBasedFilePathMapping as AnchorBasedFilePathMapping,
    CodeReferencesMetadataSet as CodeReferencesMetadataSet,
    CodeReferencesMetadataValue as CodeReferencesMetadataValue,
    FilePathMapping as FilePathMapping,
    LocalFileCodeReference as LocalFileCodeReference,
    UrlCodeReference as UrlCodeReference,
    link_code_references_to_git as link_code_references_to_git,
    with_source_code_references as with_source_code_references,
)
from dagster._core.definitions.metadata.table import (
    TableColumn as TableColumn,
    TableColumnConstraints as TableColumnConstraints,
    TableColumnDep as TableColumnDep,
    TableColumnLineage as TableColumnLineage,
    TableConstraints as TableConstraints,
    TableRecord as TableRecord,
    TableSchema as TableSchema,
)
from dagster._core.errors import DagsterInvalidMetadata
from dagster._serdes import whitelist_for_serdes
from dagster._serdes.serdes import (
    FieldSerializer,
    PackableValue,
    UnpackContext,
    WhitelistMap,
    pack_value,
)
from dagster._utils.warnings import deprecation_warning, normalize_renamed_param

ArbitraryMetadataMapping: TypeAlias = Mapping[str, Any]

RawMetadataValue = Union[
    MetadataValue,
    TableSchema,
    AssetKey,
    os.PathLike,
    dict[Any, Any],
    float,
    int,
    list[Any],
    str,
    datetime,
    None,
]

MetadataMapping: TypeAlias = Mapping[str, MetadataValue]
RawMetadataMapping: TypeAlias = Mapping[str, RawMetadataValue]

T_Packable = TypeVar("T_Packable", bound=PackableValue, default=PackableValue, covariant=True)

# ########################
# ##### NORMALIZATION
# ########################


def normalize_metadata(
    metadata: Mapping[str, RawMetadataValue],
    allow_invalid: bool = False,
) -> Mapping[str, MetadataValue]:
    # This is a stopgap measure to deal with unsupported metadata values, which occur when we try
    # to convert arbitrary metadata (on e.g. OutputDefinition) to a MetadataValue, which is required
    # for serialization. This will cause unsupported values to be silently replaced with a
    # string placeholder.
    normalized_metadata: dict[str, MetadataValue] = {}
    for k, v in metadata.items():
        try:
            normalized_value = normalize_metadata_value(v)
        except DagsterInvalidMetadata as e:
            if allow_invalid:
                deprecation_warning(
                    "Support for arbitrary metadata values",
                    "2.0.0",
                    additional_warn_text=(
                        "In the future, all user-supplied metadata values must be one of"
                        f" {RawMetadataValue}"
                    ),
                    stacklevel=4,  # to get the caller of `normalize_metadata`
                )
                normalized_value = TextMetadataValue(f"[{v.__class__.__name__}] (unserializable)")
            else:
                raise DagsterInvalidMetadata(
                    f'Could not resolve the metadata value for "{k}" to a known type. {e}'
                ) from None
        normalized_metadata[k] = normalized_value

    return normalized_metadata


def has_corresponding_metadata_value_class(obj: Any) -> bool:
    return isinstance(obj, (str, float, bool, int, list, dict, os.PathLike, AssetKey, TableSchema))


def normalize_metadata_value(raw_value: RawMetadataValue) -> "MetadataValue[Any]":
    if isinstance(raw_value, MetadataValue):
        return raw_value
    elif isinstance(raw_value, str):
        return MetadataValue.text(raw_value)
    elif isinstance(raw_value, float):
        return MetadataValue.float(raw_value)
    elif isinstance(raw_value, bool):
        return MetadataValue.bool(raw_value)
    elif isinstance(raw_value, int):
        return MetadataValue.int(raw_value)
    elif isinstance(raw_value, (list, dict)):
        return MetadataValue.json(raw_value)
    elif isinstance(raw_value, os.PathLike):
        return MetadataValue.path(raw_value)
    elif isinstance(raw_value, AssetKey):
        return MetadataValue.asset(raw_value)
    elif isinstance(raw_value, TableSchema):
        return MetadataValue.table_schema(raw_value)
    elif isinstance(raw_value, TableColumnLineage):
        return MetadataValue.column_lineage(raw_value)
    elif raw_value is None:
        return MetadataValue.null()

    raise DagsterInvalidMetadata(
        f"Its type was {type(raw_value)}. Consider wrapping the value with the appropriate "
        "MetadataValue type."
    )


# ########################
# ##### METADATA BACKCOMPAT
# ########################

# Metadata used to be represented as a `List[MetadataEntry]`, but that class has been deleted. But
# we still serialize metadata dicts to the serialized representation of `List[MetadataEntry]` for
# backcompat purposes.


class MetadataFieldSerializer(FieldSerializer):
    """Converts between metadata dict (new) and metadata entries list (old)."""

    storage_name = "metadata_entries"
    loaded_name = "metadata"

    def pack(
        self,
        metadata_dict: Mapping[str, MetadataValue],
        whitelist_map: WhitelistMap,
        descent_path: str,
    ) -> Sequence[Mapping[str, Any]]:
        return [
            {
                "__class__": "EventMetadataEntry",
                "label": k,
                # MetadataValue itself can't inherit from NamedTuple and so isn't a PackableValue,
                # but one of its subclasses will always be returned here.
                "entry_data": pack_value(v, whitelist_map, descent_path),  # type: ignore
                "description": None,
            }
            for k, v in metadata_dict.items()
        ]

    def unpack(
        self,
        metadata_entries: list["MetadataEntry"],
        whitelist_map: WhitelistMap,
        context: UnpackContext,
    ) -> Mapping[str, MetadataValue]:
        return {e.label: e.entry_data for e in metadata_entries}


T_MetadataValue = TypeVar("T_MetadataValue", bound=MetadataValue, covariant=True)


# NOTE: MetadataEntry is no longer accessible via the public API-- all metadata APIs use metadata
# dicts. This clas shas only been preserved to adhere strictly to our backcompat guarantees. It is
# still instantiated in the above `MetadataFieldSerializer` but that can easily be changed.
@deprecated(
    breaking_version="2.0",
    additional_warn_text="Please use a dict with `MetadataValue` values instead.",
)
@deprecated_param(
    param="entry_data", breaking_version="2.0", additional_warn_text="Use `value` instead."
)
@whitelist_for_serdes(storage_name="EventMetadataEntry")
class MetadataEntry(
    NamedTuple(
        "_MetadataEntry",
        [
            ("label", PublicAttr[str]),
            ("description", PublicAttr[Optional[str]]),
            ("entry_data", PublicAttr[MetadataValue]),
        ],
    ),
    Generic[T_MetadataValue],
):
    """A structure for describing metadata for Dagster events.

    .. note:: This class is no longer usable in any Dagster API, and will be completely removed in 2.0.

    Lists of objects of this type can be passed as arguments to Dagster events and will be displayed
    in the Dagster UI and other tooling.

    Should be yielded from within an IO manager to append metadata for a given input/output event.
    For other event types, passing a dict with `MetadataValue` values to the `metadata` argument
    is preferred.

    Args:
        label (str): Short display label for this metadata entry.
        description (Optional[str]): A human-readable description of this metadata entry.
        value (MetadataValue): Typed metadata entry data. The different types allow
            for customized display in tools like the Dagster UI.
    """

    def __new__(
        cls,
        label: str,
        description: Optional[str] = None,
        entry_data: Optional["RawMetadataValue"] = None,
        value: Optional["RawMetadataValue"] = None,
    ):
        value = cast(
            RawMetadataValue,
            normalize_renamed_param(
                new_val=value,
                new_arg="value",
                old_val=entry_data,
                old_arg="entry_data",
            ),
        )
        value = normalize_metadata_value(value)

        return super().__new__(
            cls,
            check.str_param(label, "label"),
            check.opt_str_param(description, "description"),
            check.inst_param(value, "value", MetadataValue),
        )

    @property
    def value(self):
        """Alias of `entry_data`."""
        return self.entry_data



import inspect
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Annotated, Any, Callable, Final, Optional, TypeVar, Union, overload

from typing_extensions import TypeAlias

from dagster import _check as check
from dagster._core.decorator_utils import (
    apply_pre_call_decorator,
    get_decorator_target,
    is_resource_def,
)
from dagster._utils.warnings import (
    beta_warning,
    deprecation_warning,
    preview_warning,
    supersession_warning,
)

# For the time being, `Annotatable` is set to `Any` even though it should be set to `Decoratable` to
# avoid choking the type checker. Choking happens because of a niche scenario where
# `ResourceDefinition`, which is part of `Decoratable`, is used as an argument to a function that
# accepts `Annotatable`. There is a certain circularity here that current versions of pyright fail
# on. It will likely be resolved by future versions of pyright, and then `Annotatable` should be set
# to `Decoratable`.
Annotatable: TypeAlias = Any

T_Annotatable = TypeVar("T_Annotatable", bound=Annotatable)


# ########################
# ##### PUBLIC
# ########################

_PUBLIC_ATTR_NAME: Final[str] = "_is_public"


def public(obj: T_Annotatable) -> T_Annotatable:
    """Mark a method on a public class as public. This distinguishes the method from "internal"
    methods, which are methods that are public in the Python sense of being non-underscored, but
    not intended for user access. Only `public` methods of a class are rendered in the docs.
    """
    target = _get_annotation_target(obj)
    setattr(target, _PUBLIC_ATTR_NAME, True)
    return obj


def is_public(obj: Annotatable) -> bool:
    target = _get_annotation_target(obj)
    return hasattr(target, _PUBLIC_ATTR_NAME) and getattr(target, _PUBLIC_ATTR_NAME)


# Use `PublicAttr` to annotate public attributes on `NamedTuple`:
#
# from dagster._annotations import PublicAttr
#
# class Foo(NamedTuple("_Foo", [("bar", PublicAttr[int])])):
#     ...

T = TypeVar("T")

PUBLIC: Final[str] = "public"

PublicAttr: TypeAlias = Annotated[T, PUBLIC]

# ########################
# ##### PREVIEW
# ########################


_PREVIEW_ATTR_NAME: Final[str] = "_preview"


@dataclass
class PreviewInfo:
    additional_warn_text: Optional[str] = None
    subject: Optional[str] = None


@overload
def preview(
    __obj: T_Annotatable,
    *,
    additional_warn_text: Optional[str] = ...,
    subject: Optional[str] = ...,
    emit_runtime_warning: bool = ...,
) -> T_Annotatable: ...


@overload
def preview(
    __obj: None = ...,
    *,
    additional_warn_text: Optional[str] = ...,
    subject: Optional[str] = ...,
    emit_runtime_warning: bool = ...,
) -> Callable[[T_Annotatable], T_Annotatable]: ...


def preview(
    __obj: Optional[T_Annotatable] = None,
    *,
    additional_warn_text: Optional[str] = None,
    subject: Optional[str] = None,
    emit_runtime_warning: bool = True,
) -> Union[T_Annotatable, Callable[[T_Annotatable], T_Annotatable]]:
    """Mark an object as preview. This appends some metadata to the object that causes it to be
    rendered with a "preview" tag and associated warning in the docs.

    If `emit_runtime_warning` is True, a warning will also be emitted when the function is called,
    having the same text as is displayed in the docs. For consistency between docs and runtime
    warnings, this decorator is preferred to manual calls to `preview_warning`.

    Args:
        additional_warn_text (Optional[str]): Additional text to display after the preview warning.
        subject (Optional[str]): The subject of the preview warning. Defaults to a string
            representation of the decorated object. This is useful when marking usage of
            a preview API inside an otherwise non-preview function, so
            that it can be easily cleaned up later. It should only be used with
            `emit_runtime_warning=False`, as we don't want to warn users when a
            preview API is used internally.
        emit_runtime_warning (bool): Whether to emit a warning when the function is called.

    Usage:

        .. code-block:: python

            @preview
            def my_preview_function(my_arg):
                ...

            @preview
            class MyPreviewClass:
                ...

            @preview(subject="some_preview_function", emit_runtime_warning=False)
            def not_preview_function():
                ...
                some_preview_function()
                ...
    """
    if __obj is None:
        return lambda obj: preview(
            obj,
            subject=subject,
            emit_runtime_warning=emit_runtime_warning,
            additional_warn_text=additional_warn_text,
        )
    else:
        target = _get_annotation_target(__obj)
        setattr(
            target,
            _PREVIEW_ATTR_NAME,
            PreviewInfo(additional_warn_text, subject),
        )

        if emit_runtime_warning:
            stack_level = _get_warning_stacklevel(__obj)
            subject = subject or _get_subject(__obj)
            warning_fn = lambda: preview_warning(
                subject,
                additional_warn_text=additional_warn_text,
                stacklevel=stack_level,
            )
            return apply_pre_call_decorator(__obj, warning_fn)
        else:
            return __obj


def is_preview(obj: Annotatable) -> bool:
    target = _get_annotation_target(obj)
    return hasattr(target, _PREVIEW_ATTR_NAME)


def get_preview_info(obj: Annotatable) -> PreviewInfo:
    target = _get_annotation_target(obj)
    return getattr(target, _PREVIEW_ATTR_NAME)


# ########################
# ##### BETA
# ########################


_BETA_ATTR_NAME: Final[str] = "_beta"


@dataclass
class BetaInfo:
    additional_warn_text: Optional[str] = None
    subject: Optional[str] = None


@overload
def beta(
    __obj: T_Annotatable,
    *,
    additional_warn_text: Optional[str] = ...,
    subject: Optional[str] = ...,
    emit_runtime_warning: bool = ...,
) -> T_Annotatable: ...


@overload
def beta(
    __obj: None = ...,
    *,
    additional_warn_text: Optional[str] = ...,
    subject: Optional[str] = ...,
    emit_runtime_warning: bool = ...,
) -> Callable[[T_Annotatable], T_Annotatable]: ...


def beta(
    __obj: Optional[T_Annotatable] = None,
    *,
    additional_warn_text: Optional[str] = None,
    subject: Optional[str] = None,
    emit_runtime_warning: bool = True,
) -> Union[T_Annotatable, Callable[[T_Annotatable], T_Annotatable]]:
    """Mark an object as beta. This appends some metadata to the object that causes it to be
    rendered with a "beta" tag and associated warning in the docs.

    If `emit_runtime_warning` is True, a warning will also be emitted when the function is called,
    having the same text as is displayed in the docs. For consistency between docs and runtime
    warnings, this decorator is preferred to manual calls to `beta_warning`.

    Args:
        additional_warn_text (Optional[str]): Additional text to display after the beta warning.
        subject (Optional[str]): The subject of the beta warning. Defaults to a string
            representation of the decorated object. This is useful when marking usage of
            a beta API inside an otherwise non-beta function, so
            that it can be easily cleaned up later. It should only be used with
            `emit_runtime_warning=False`, as we don't want to warn users when a
            beta API is used internally.
        emit_runtime_warning (bool): Whether to emit a warning when the function is called.

    Usage:

        .. code-block:: python

            @beta
            def my_beta_function(my_arg):
                ...

            @beta
            class MyBetaClass:
                ...

            @beta(subject="some_beta_function", emit_runtime_warning=False)
            def not_beta_function():
                ...
                some_beta_function()
                ...
    """
    if __obj is None:
        return lambda obj: beta(
            obj,
            subject=subject,
            emit_runtime_warning=emit_runtime_warning,
            additional_warn_text=additional_warn_text,
        )
    else:
        target = _get_annotation_target(__obj)
        setattr(
            target,
            _BETA_ATTR_NAME,
            BetaInfo(additional_warn_text, subject),
        )

        if emit_runtime_warning:
            stack_level = _get_warning_stacklevel(__obj)
            subject = subject or _get_subject(__obj)
            warning_fn = lambda: beta_warning(
                subject,
                additional_warn_text=additional_warn_text,
                stacklevel=stack_level,
            )
            return apply_pre_call_decorator(__obj, warning_fn)
        else:
            return __obj


def is_beta(obj: Annotatable) -> bool:
    target = _get_annotation_target(obj)
    return hasattr(target, _BETA_ATTR_NAME)


def get_beta_info(obj: Annotatable) -> BetaInfo:
    target = _get_annotation_target(obj)
    return getattr(target, _BETA_ATTR_NAME)


# ########################
# ##### BETA PARAM
# ########################

_BETA_PARAM_ATTR_NAME: Final[str] = "_beta_params"


@overload
def beta_param(
    __obj: T_Annotatable,
    *,
    param: str,
    additional_warn_text: Optional[str] = ...,
    emit_runtime_warning: bool = ...,
) -> T_Annotatable: ...


@overload
def beta_param(
    __obj: None = ...,
    *,
    param: str,
    additional_warn_text: Optional[str] = ...,
    emit_runtime_warning: bool = ...,
) -> Callable[[T_Annotatable], T_Annotatable]: ...


def beta_param(
    __obj: Optional[T_Annotatable] = None,
    *,
    param: str,
    additional_warn_text: Optional[str] = None,
    emit_runtime_warning: bool = True,
) -> Union[T_Annotatable, Callable[[T_Annotatable], T_Annotatable]]:
    """Mark a parameter of a class initializer or function/method as beta. This appends some
    metadata to the decorated object that causes the specified argument to be rendered with a
    "beta" tag and associated warning in the docs.

    If `emit_runtime_warning` is True, a warning will also be emitted when the function is called
    and a non-None value is passed for the parameter. For consistency between docs and runtime
    warnings, this decorator is preferred to manual calls to `beta_warning`. Note that the
    warning will only be emitted if the value is passed as a keyword argument.

    Args:
        param (str): The name of the parameter to mark as beta.
        additional_warn_text (str): Additional text to display after the beta warning.
            Typically, this should suggest a newer API.
        emit_runtime_warning (bool): Whether to emit a warning when the function is called.
    """
    if __obj is None:
        return lambda obj: beta_param(
            obj,
            param=param,
            additional_warn_text=additional_warn_text,
            emit_runtime_warning=emit_runtime_warning,
        )
    else:
        check.invariant(
            _annotatable_has_param(__obj, param),
            f"Attempted to mark undefined parameter `{param}` beta.",
        )
        target = _get_annotation_target(__obj)

        if not hasattr(target, _BETA_PARAM_ATTR_NAME):
            setattr(target, _BETA_PARAM_ATTR_NAME, {})
        getattr(target, _BETA_PARAM_ATTR_NAME)[param] = BetaInfo(
            additional_warn_text=additional_warn_text,
        )

        if emit_runtime_warning:
            condition = lambda *_, **kwargs: kwargs.get(param) is not None
            warning_fn = lambda: beta_warning(
                _get_subject(__obj, param=param),
                additional_warn_text=additional_warn_text,
                stacklevel=4,
            )
            return apply_pre_call_decorator(__obj, warning_fn, condition=condition)
        else:
            return __obj


def has_beta_params(obj: Annotatable) -> bool:
    return hasattr(_get_annotation_target(obj), _BETA_PARAM_ATTR_NAME)


def get_beta_params(obj: Annotatable) -> Mapping[str, BetaInfo]:
    return getattr(_get_annotation_target(obj), _BETA_PARAM_ATTR_NAME)


def is_beta_param(obj: Annotatable, param_name: str) -> bool:
    target = _get_annotation_target(obj)
    return param_name in getattr(target, _BETA_PARAM_ATTR_NAME, {})


def get_beta_param_info(obj: Annotatable, param_name: str) -> BetaInfo:
    target = _get_annotation_target(obj)
    return getattr(target, _BETA_PARAM_ATTR_NAME)[param_name]


# ########################
# ##### SUPERSEDED
# ########################


_SUPERSEDED_ATTR_NAME: Final[str] = "_superseded"


@dataclass
class SupersededInfo:
    additional_warn_text: Optional[str] = None
    subject: Optional[str] = None


@overload
def superseded(
    __obj: T_Annotatable,
    *,
    additional_warn_text: Optional[str] = ...,
    subject: Optional[str] = ...,
    emit_runtime_warning: bool = ...,
) -> T_Annotatable: ...


@overload
def superseded(
    __obj: None = ...,
    *,
    additional_warn_text: Optional[str] = ...,
    subject: Optional[str] = ...,
    emit_runtime_warning: bool = ...,
) -> Callable[[T_Annotatable], T_Annotatable]: ...


def superseded(
    __obj: Optional[T_Annotatable] = None,
    *,
    additional_warn_text: Optional[str] = None,
    subject: Optional[str] = None,
    emit_runtime_warning: bool = True,
) -> Union[T_Annotatable, Callable[[T_Annotatable], T_Annotatable]]:
    # TODO: add "superseded" warning to docs
    """Mark an object as superseded. This appends some metadata to the object that causes it to be
    rendered with a "superseded" tag and associated warning in the docs.

    If `emit_runtime_warning` is True, a warning will also be emitted when the function is called,
    having the same text as is displayed in the docs. For consistency between docs and runtime
    warnings, this decorator is preferred to manual calls to `supersession_warning`.

    Args:
        additional_warn_text (Optional[str]): Additional text to display after the supersession warning.
            Typically this should suggest a newer API.
        subject (Optional[str]): The subject of the supersession warning. Defaults to a string
            representation of the decorated object. This is useful when marking usage of
            a superseded API inside an otherwise non-superseded function, so
            that it can be easily cleaned up later. It should only be used with
            `emit_runtime_warning=False`, as we don't want to warn users when a
            superseded API is used internally.
        emit_runtime_warning (bool): Whether to emit a warning when the function is called.

    Usage:

        .. code-block:: python

            @superseded(additional_warn_text="Use my_new_function instead")
            def my_superseded_function(my_arg):
                ...

            @superseded(additional_warn_text="Use MyNewClass instead")
            class MySupersededClass:
                ...

            @superseded(subject="some_superseded_function", emit_runtime_warning=False)
            def not_superseded_function():
                ...
                some_superseded_function()
                ...
    """
    if __obj is None:
        return lambda obj: superseded(
            obj,
            subject=subject,
            emit_runtime_warning=emit_runtime_warning,
            additional_warn_text=additional_warn_text,
        )
    else:
        target = _get_annotation_target(__obj)
        setattr(
            target,
            _SUPERSEDED_ATTR_NAME,
            SupersededInfo(additional_warn_text, subject),
        )

        if emit_runtime_warning:
            stack_level = _get_warning_stacklevel(__obj)
            subject = subject or _get_subject(__obj)
            warning_fn = lambda: supersession_warning(
                subject,
                additional_warn_text=additional_warn_text,
                stacklevel=stack_level,
            )
            return apply_pre_call_decorator(__obj, warning_fn)
        else:
            return __obj


def is_superseded(obj: Annotatable) -> bool:
    target = _get_annotation_target(obj)
    return hasattr(target, _SUPERSEDED_ATTR_NAME)


def get_superseded_info(obj: Annotatable) -> SupersededInfo:
    target = _get_annotation_target(obj)
    return getattr(target, _SUPERSEDED_ATTR_NAME)


# ########################
# ##### DEPRECATED
# ########################


_DEPRECATED_ATTR_NAME: Final[str] = "_deprecated"


@dataclass
class DeprecatedInfo:
    breaking_version: str
    hidden: bool
    additional_warn_text: Optional[str]
    subject: Optional[str]


@overload
def deprecated(
    __obj: T_Annotatable,
    *,
    breaking_version: str,
    additional_warn_text: Optional[str] = ...,
    subject: Optional[str] = ...,
    emit_runtime_warning: bool = ...,
) -> T_Annotatable: ...


@overload
def deprecated(
    __obj: None = ...,
    *,
    breaking_version: str,
    additional_warn_text: Optional[str] = ...,
    subject: Optional[str] = ...,
    emit_runtime_warning: bool = ...,
) -> Callable[[T_Annotatable], T_Annotatable]: ...


def deprecated(
    __obj: Optional[T_Annotatable] = None,
    *,
    breaking_version: str,
    additional_warn_text: Optional[str] = None,
    subject: Optional[str] = None,
    emit_runtime_warning: bool = True,
) -> Union[T_Annotatable, Callable[[T_Annotatable], T_Annotatable]]:
    """Mark an object as deprecated. This appends some metadata to the object that causes it to be
    rendered with a "deprecated" tag and associated warning in the docs.

    If `emit_runtime_warning` is True, a warning will also be emitted when the function is called,
    having the same text as is displayed in the docs. For consistency between docs and runtime
    warnings, this decorator is preferred to manual calls to `deprecation_warning`.

    Args:
        breaking_version (str): The version at which the deprecated function will be removed.
        additional_warn_text (Optional[str]): Additional text to display after the deprecation warning.
            Typically this should suggest a newer API.
        subject (Optional[str]): The subject of the deprecation warning. Defaults to a string
            representation of the decorated object. This is useful when marking usage of
            a deprecated API inside an otherwise non-deprecated function, so
            that it can be easily cleaned up later. It should only be used with
            `emit_runtime_warning=False`, as we don't want to warn users when a
            deprecated API is used internally.
        emit_runtime_warning (bool): Whether to emit a warning when the function is called.

    Usage:
        .. code-block:: python

            @deprecated(breaking_version="2.0", additional_warn_text="Use my_new_function instead")
            def my_deprecated_function(my_arg):
                ...

            @deprecated(breaking_version="2.0", additional_warn_text="Use MyNewClass instead")
            class MyDeprecatedClass:
                ...

            @deprecated(breaking_version="2.0", subject="some_deprecated_function", emit_runtime_warning=False)
            def not_deprecated_function():
                ...
                some_deprecated_function()
                ...

    """
    if __obj is None:
        return lambda obj: deprecated(
            obj,
            subject=subject,
            emit_runtime_warning=emit_runtime_warning,
            breaking_version=breaking_version,
            additional_warn_text=additional_warn_text,
        )
    else:
        target = _get_annotation_target(__obj)
        setattr(
            target,
            _DEPRECATED_ATTR_NAME,
            DeprecatedInfo(
                breaking_version=breaking_version,
                additional_warn_text=additional_warn_text,
                subject=subject,
                hidden=False,
            ),
        )

        if emit_runtime_warning:
            stack_level = _get_warning_stacklevel(__obj)
            subject = subject or _get_subject(__obj)
            warning_fn = lambda: deprecation_warning(
                subject,
                breaking_version=breaking_version,
                additional_warn_text=additional_warn_text,
                stacklevel=stack_level,
            )
            return apply_pre_call_decorator(__obj, warning_fn)
        else:
            return __obj


def is_deprecated(obj: Annotatable) -> bool:
    target = _get_annotation_target(obj)
    return hasattr(target, _DEPRECATED_ATTR_NAME)


def get_deprecated_info(obj: Annotatable) -> DeprecatedInfo:
    target = _get_annotation_target(obj)
    return getattr(target, _DEPRECATED_ATTR_NAME)


# ########################
# ##### DEPRECATED PARAM
# ########################

_DEPRECATED_PARAM_ATTR_NAME: Final[str] = "_deprecated_params"


@overload
def deprecated_param(
    __obj: T_Annotatable,
    *,
    param: str,
    breaking_version: str,
    additional_warn_text: Optional[str] = ...,
    emit_runtime_warning: bool = ...,
) -> T_Annotatable: ...


@overload
def deprecated_param(
    __obj: None = ...,
    *,
    param: str,
    breaking_version: str,
    additional_warn_text: Optional[str] = ...,
    emit_runtime_warning: bool = ...,
) -> Callable[[T_Annotatable], T_Annotatable]: ...


def deprecated_param(
    __obj: Optional[T_Annotatable] = None,
    *,
    param: str,
    breaking_version: str,
    additional_warn_text: Optional[str] = None,
    emit_runtime_warning: bool = True,
) -> T_Annotatable:
    """Mark a parameter of a class initializer or function/method as deprecated. This appends some
    metadata to the decorated object that causes the specified argument to be rendered with a
    "deprecated" tag and associated warning in the docs.

    If `emit_runtime_warning` is True, a warning will also be emitted when the function is called
    and a non-None value is passed for the parameter. For consistency between docs and runtime
    warnings, this decorator is preferred to manual calls to `deprecation_warning`. Note that the
    warning will only be emitted if the value is passed as a keyword argument.

    Args:
        param (str): The name of the parameter to deprecate.
        breaking_version (str): The version at which the deprecated function will be removed.
        additional_warn_text (str): Additional text to display after the deprecation warning.
            Typically this should suggest a newer API.
        emit_runtime_warning (bool): Whether to emit a warning when the function is called.
        hidden (bool): Whether or not this is a hidden parameters. Hidden parameters are only
            passed via kwargs and are hidden from the type signature. This makes it so
            that this hidden parameter does not appear in typeaheads. In order to provide
            high quality error messages we also provide the helper function
            only_allow_hidden_params_in_kwargs to ensure there are high quality
            error messages if the user passes an unsupported keyword argument.


    """
    if __obj is None:
        return lambda obj: deprecated_param(  # type: ignore
            obj,
            param=param,
            breaking_version=breaking_version,
            additional_warn_text=additional_warn_text,
            emit_runtime_warning=emit_runtime_warning,
        )
    else:
        return attach_deprecation_info_and_wrap(
            __obj,
            param=param,
            breaking_version=breaking_version,
            additional_warn_text=additional_warn_text,
            emit_runtime_warning=emit_runtime_warning,
            hidden=False,
        )


def attach_deprecation_info_and_wrap(
    obj: T_Annotatable,
    param: str,
    breaking_version: str,
    additional_warn_text: Optional[str] = None,
    emit_runtime_warning: bool = True,
    hidden: bool = False,
) -> T_Annotatable:
    if not hidden:
        check.invariant(
            _annotatable_has_param(obj, param),
            f"Attempted to mark undefined parameter `{param}` deprecated.",
        )
    target = _get_annotation_target(obj)
    if not hasattr(target, _DEPRECATED_PARAM_ATTR_NAME):
        setattr(target, _DEPRECATED_PARAM_ATTR_NAME, {})
    getattr(target, _DEPRECATED_PARAM_ATTR_NAME)[param] = DeprecatedInfo(
        breaking_version=breaking_version,
        additional_warn_text=additional_warn_text,
        hidden=hidden,
        subject=None,
    )

    if not emit_runtime_warning:
        return obj

    condition = lambda *_, **kwargs: kwargs.get(param) is not None
    warning_fn = lambda: deprecation_warning(
        _get_subject(obj, param=param),
        breaking_version=breaking_version,
        additional_warn_text=additional_warn_text,
        stacklevel=4,
    )
    return apply_pre_call_decorator(obj, warning_fn, condition=condition)


@overload
def hidden_param(
    __obj: T_Annotatable,
    *,
    param: str,
    breaking_version: str,
    additional_warn_text: Optional[str] = ...,
    emit_runtime_warning: bool = ...,
) -> T_Annotatable: ...


@overload
def hidden_param(
    __obj: None = ...,
    *,
    param: str,
    breaking_version: str,
    additional_warn_text: Optional[str] = ...,
    emit_runtime_warning: bool = ...,
) -> Callable[[T_Annotatable], T_Annotatable]: ...


def hidden_param(
    __obj: Optional[T_Annotatable] = None,
    *,
    param: str,
    breaking_version: str,
    additional_warn_text: Optional[str] = None,
    emit_runtime_warning: bool = True,
) -> T_Annotatable:
    """Hidden parameters are only passed via kwargs and are hidden from the
    type signature. This makes it so that this hidden parameter does not
    appear in typeaheads. In order to provide high quality error messages
    we also provide the helper function only_allow_hidden_params_in_kwargs
    to ensure there are high quality error messages if the user passes
    an unsupported keyword argument.

    Args:
        breaking_version (str): The version at which the deprecated function will be removed.
        additional_warn_text (Optional[str]): Additional text to display after the deprecation warning.
            Typically this should suggest a newer API.
        subject (Optional[str]): The subject of the deprecation warning. Defaults to a string
            representation of the decorated object. This is useful when marking usage of
            a deprecated API inside an otherwise non-deprecated function, so
            that it can be easily cleaned up later. It should only be used with
            `emit_runtime_warning=False`, as we don't want to warn users when a
            deprecated API is used internally.
        emit_runtime_warning (bool): Whether to emit a warning when the function is called.

    Usage:
        .. code-block:: python

            @hidden_param(breaking_version="2.0", additional_warn_text="Use my_new_function instead")
            def func_with_hidden_args(**kwargs):
                only_allow_hidden_params_in_kwargs(func_with_hidden_args, kwargs)
    """
    if __obj is None:
        return lambda obj: hidden_param(  # type: ignore
            obj,
            param=param,
            breaking_version=breaking_version,
            additional_warn_text=additional_warn_text,
            emit_runtime_warning=emit_runtime_warning,
        )
    else:
        return attach_deprecation_info_and_wrap(
            __obj,
            param=param,
            breaking_version=breaking_version,
            additional_warn_text=additional_warn_text,
            emit_runtime_warning=emit_runtime_warning,
            hidden=True,
        )


def has_deprecated_params(obj: Annotatable) -> bool:
    return hasattr(_get_annotation_target(obj), _DEPRECATED_PARAM_ATTR_NAME)


def get_deprecated_params(obj: Annotatable) -> Mapping[str, DeprecatedInfo]:
    return getattr(_get_annotation_target(obj), _DEPRECATED_PARAM_ATTR_NAME)


def is_deprecated_param(obj: Annotatable, param_name: str) -> bool:
    target = _get_annotation_target(obj)
    return param_name in getattr(target, _DEPRECATED_PARAM_ATTR_NAME, {})


def get_deprecated_param_info(obj: Annotatable, param_name: str) -> DeprecatedInfo:
    target = _get_annotation_target(obj)
    return getattr(target, _DEPRECATED_PARAM_ATTR_NAME)[param_name]


# ########################
# ##### HELPERS
# ########################


def copy_annotations(dest: Annotatable, src: Annotatable) -> None:
    """Copy all Dagster annotations from one object to another object."""
    dest_target = _get_annotation_target(dest)
    src_target = _get_annotation_target(src)
    if hasattr(src_target, _PUBLIC_ATTR_NAME):
        setattr(dest_target, _PUBLIC_ATTR_NAME, getattr(src_target, _PUBLIC_ATTR_NAME))
    if hasattr(src_target, _DEPRECATED_ATTR_NAME):
        setattr(dest_target, _DEPRECATED_ATTR_NAME, getattr(src_target, _DEPRECATED_ATTR_NAME))
    if hasattr(src_target, _DEPRECATED_PARAM_ATTR_NAME):
        setattr(
            dest_target,
            _DEPRECATED_PARAM_ATTR_NAME,
            getattr(src_target, _DEPRECATED_PARAM_ATTR_NAME),
        )
    if hasattr(src_target, _SUPERSEDED_ATTR_NAME):
        setattr(dest_target, _SUPERSEDED_ATTR_NAME, getattr(src_target, _SUPERSEDED_ATTR_NAME))
    if hasattr(src_target, _PREVIEW_ATTR_NAME):
        setattr(dest_target, _PREVIEW_ATTR_NAME, getattr(src_target, _PREVIEW_ATTR_NAME))
    if hasattr(src_target, _BETA_ATTR_NAME):
        setattr(dest_target, _BETA_ATTR_NAME, getattr(src_target, _BETA_ATTR_NAME))
    if hasattr(src_target, _BETA_PARAM_ATTR_NAME):
        setattr(
            dest_target,
            _BETA_PARAM_ATTR_NAME,
            getattr(src_target, _BETA_PARAM_ATTR_NAME),
        )


def _get_annotation_target(obj: Annotatable) -> object:
    """Given an object to be annotated, return the underlying object that will actually store the annotations.
    This is necessary because not all objects are mutable, and so can't be annotated directly.
    """
    if isinstance(obj, property):
        return obj.fget
    elif isinstance(obj, (staticmethod, classmethod)):
        return obj.__func__
    else:
        return obj


def _get_subject(obj: Annotatable, param: Optional[str] = None) -> str:
    """Get the string representation of an annotated object that will appear in
    annotation-generated warnings about the object.
    """
    if param:
        if isinstance(obj, type):
            return f"Parameter `{param}` of initializer `{obj.__qualname__}.__init__`"
        else:
            fn_subject = _get_subject(obj)
            return f"Parameter `{param}` of {fn_subject[:1].lower() + fn_subject[1:]}"
    else:
        if isinstance(obj, type):
            return f"Class `{obj.__qualname__}`"
        elif isinstance(obj, property):
            return f"Property `{obj.fget.__qualname__ if obj.fget else obj}`"
        # classmethod and staticmethod don't themselves get a `__qualname__` attr until Python 3.10.
        elif isinstance(obj, classmethod):
            return f"Class method `{_get_annotation_target(obj).__qualname__}`"
        elif isinstance(obj, staticmethod):
            return f"Static method `{_get_annotation_target(obj).__qualname__}`"
        elif inspect.isfunction(obj):
            return f"Function `{obj.__qualname__}`"
        elif is_resource_def(obj):
            return f"Dagster resource `{obj.__qualname__}`"
        else:
            check.failed(f"Unexpected object type: {type(obj)}")


def _get_warning_stacklevel(obj: Annotatable):
    """Get the stacklevel to use for warnings that are attached to a target via decorator.

    The goal is to have the warning point to the line where the function in the
    underlying object is actually invoked. This isn't straightforward
    because some objects have complicated logic in between `__call__` and
    the site at which a wrapped function containing the warning is actually
    called. Can be determined through trial and error.
    """
    if is_resource_def(obj):
        return 6
    else:
        return 4


def _annotatable_has_param(obj: Annotatable, param: str) -> bool:
    target_fn = get_decorator_target(obj)
    return param in inspect.signature(target_fn).parameters


def only_allow_hidden_params_in_kwargs(annotatable: Annotatable, kwargs: Mapping[str, Any]) -> None:
    deprecated_params = (
        get_deprecated_params(annotatable) if has_deprecated_params(annotatable) else {}
    )
    for param in kwargs:
        if param not in deprecated_params:
            raise TypeError(f"{annotatable.__name__} got an unexpected keyword argument '{param}'")

        check.invariant(
            deprecated_params[param].hidden,
            f"Unexpected non-hidden deprecated parameter '{param}' in kwargs. Should never get here.",
        )



"""0.10.0 create new event log tables.

Revision ID: 942138e33bf9
Revises: a0234163e0e3
Create Date: 2021-01-13 12:54:27.921898

"""

from dagster._core.storage.migration.utils import create_0_10_0_event_log_tables

# revision identifiers, used by Alembic.
revision = "942138e33bf9"
down_revision = "a0234163e0e3"
branch_labels = None
depends_on = None


def upgrade():
    create_0_10_0_event_log_tables()


def downgrade():
    pass



import re
import warnings
from collections import defaultdict
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, Optional, Union

import dagster._seven as seven
from dagster import _check as check
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._core.storage.tags import SYSTEM_TAG_PREFIX, USER_EDITABLE_SYSTEM_TAGS

if TYPE_CHECKING:
    from dagster._core.execution.plan.step import ExecutionStep
    from dagster._core.storage.dagster_run import DagsterRun


class TagConcurrencyLimitsCounter:
    """Helper object that keeps track of when the tag concurrency limits are met."""

    _key_limits: dict[str, int]
    _key_value_limits: dict[tuple[str, str], int]
    _unique_value_limits: dict[str, int]
    _key_counts: dict[str, int]
    _key_value_counts: dict[tuple[str, str], int]
    _unique_value_counts: dict[tuple[str, str], int]

    def __init__(
        self,
        tag_concurrency_limits: Sequence[Mapping[str, Any]],
        in_progress_tagged_items: Sequence[Union["DagsterRun", "ExecutionStep"]],
    ):
        check.opt_list_param(tag_concurrency_limits, "tag_concurrency_limits", of_type=dict)
        check.list_param(in_progress_tagged_items, "in_progress_tagged_items")

        self._key_limits = {}
        self._key_value_limits = {}
        self._unique_value_limits = {}

        for tag_limit in tag_concurrency_limits:
            key = tag_limit["key"]
            value = tag_limit.get("value")
            limit = tag_limit["limit"]

            if isinstance(value, str):
                self._key_value_limits[(key, value)] = limit
            elif not value or not value["applyLimitPerUniqueValue"]:
                self._key_limits[key] = limit
            else:
                self._unique_value_limits[key] = limit

        self._key_counts = defaultdict(lambda: 0)
        self._key_value_counts = defaultdict(lambda: 0)
        self._unique_value_counts = defaultdict(lambda: 0)

        # initialize counters based on current in progress item
        for item in in_progress_tagged_items:
            self.update_counters_with_launched_item(item)

    def is_blocked(self, item: Union["DagsterRun", "ExecutionStep"]) -> bool:
        """True if there are in progress item which are blocking this item based on tag limits."""
        for key, value in item.tags.items():
            if key in self._key_limits and self._key_counts[key] >= self._key_limits[key]:
                return True

            tag_tuple = (key, value)
            if (
                tag_tuple in self._key_value_limits
                and self._key_value_counts[tag_tuple] >= self._key_value_limits[tag_tuple]
            ):
                return True

            if (
                key in self._unique_value_limits
                and self._unique_value_counts[tag_tuple] >= self._unique_value_limits[key]
            ):
                return True

        return False

    def update_counters_with_launched_item(
        self, item: Union["DagsterRun", "ExecutionStep"]
    ) -> None:
        """Add a new in progress item to the counters."""
        for key, value in item.tags.items():
            if key in self._key_limits:
                self._key_counts[key] += 1

            tag_tuple = (key, value)
            if tag_tuple in self._key_value_limits:
                self._key_value_counts[tag_tuple] += 1

            if key in self._unique_value_limits:
                self._unique_value_counts[tag_tuple] += 1


def get_boolean_tag_value(tag_value: Optional[str], default_value: bool = False) -> bool:
    if tag_value is None:
        return default_value

    return tag_value.lower() not in {"false", "none", "0", ""}


# ########################
# ##### NORMALIZATION
# ########################

# Tag key constraints are inspired by allowed Kubernetes labels:
# https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/#syntax-and-character-set

# We allow in some cases for users to specify multi-level namespaces for tags,
# right now we only allow this for the `dagster/kind` namespace, which is how asset kinds are
# encoded under the hood.
VALID_NESTED_NAMESPACES_TAG_KEYS = r"dagster/kind/"
VALID_TAG_KEY_REGEX = re.compile(
    r"^([A-Za-z0-9_.-]{1,63}/|" + VALID_NESTED_NAMESPACES_TAG_KEYS + r")?[A-Za-z0-9_.-]{1,63}$"
)
VALID_TAG_KEY_EXPLANATION = (
    "Allowed characters: alpha-numeric, '_', '-', '.'. "
    "Tag keys can also contain a namespace section, separated by a '/'. Each section "
    "must have <= 63 characters."
)

VALID_STRICT_TAG_VALUE_REGEX = re.compile(r"^[A-Za-z0-9_.-]{0,63}$")


def normalize_tags(
    tags: Optional[Mapping[str, Any]],
    strict: bool = False,
    allow_private_system_tags: bool = True,
    warning_stacklevel: int = 4,
) -> Mapping[str, str]:
    """Normalizes key-value tags attached to definitions throughout Dagster.

    Tag normalization is complicated for backcompat reasons. In the past, tags were permitted to be
    arbitrary and potentially large JSON-serializable objects. This is inconsistent with the vision
    we have for tags going forward, which is as short string labels used for filtering and grouping
    in the UI.

    The `strict` flag controls whether to normalize/validate tags according to the new vision or the
    old. `strict` should be set whenever we are normalizing a tags parameter that is newly added. It
    should not be set if we are normalizing an older tags parameter for which we are maintaining old
    behavior.

    Args:
        strict (bool):
            If `strict=True`, we accept a restricted character set and impose length restrictions
            (<=63 characters) for string keys and values. Violations of these constraints raise
            errors. If `strict=False` then we run the same test but only warn for keys. Values are
            permitted to be any JSON-serializable object that is unaffected by JSON round-trip.
            Unserializable or round-trip-unequal values raise errors. Values are normalized to the
            JSON string representation in the return value.
        allow_private_system_tags (bool):
            Whether to allow non-whitelisted tags that start with the system tag prefix. This should
            be set to False whenever we are dealing with exclusively user-provided tags.
        warning_stacklevel (int):
            The stacklevel to use for warnings. This should be set to the calling function's
            stacklevel.

    Returns:
        Mapping[str, str]: A dictionary of normalized tags.
    """
    normalized_tags: dict[str, str] = {}
    invalid_tag_keys = []

    for key, value in check.opt_mapping_param(tags, "tags", key_type=str).items():
        # Validate the key
        if not isinstance(key, str):
            raise DagsterInvalidDefinitionError("Tag keys must be strings")
        elif (not allow_private_system_tags) and is_private_system_tag_key(key):
            raise DagsterInvalidDefinitionError(
                f"Attempted to set tag with reserved system prefix: {key}"
            )
        elif not is_valid_tag_key(key):
            invalid_tag_keys.append(key)

        # Normalize the value
        if not isinstance(value, str):
            if strict:
                raise DagsterInvalidDefinitionError("Tag values must be strings")
            else:
                normalized_tags[key] = _normalize_value(value, key)
        else:
            if strict and not is_valid_strict_tag_value(value):
                raise DagsterInvalidDefinitionError(
                    f"Invalid tag value: {value}, for key: {key}. Allowed characters: alpha-numeric, '_', '-', '.'. "
                    "Must have <= 63 characters."
                )
            normalized_tags[key] = value

    # Issue errors (strict=True) or warnings (strict=False) for any invalid tag keys that are too
    # long or contain invalid characters.
    if invalid_tag_keys:
        invalid_tag_keys_sample = invalid_tag_keys[: min(5, len(invalid_tag_keys))]
        if strict:
            raise DagsterInvalidDefinitionError(
                f"Found invalid tag keys: {invalid_tag_keys_sample}. {VALID_TAG_KEY_EXPLANATION}"
            )
        else:
            warnings.warn(
                f"Non-compliant tag keys like {invalid_tag_keys_sample} are deprecated. {VALID_TAG_KEY_EXPLANATION}",
                category=DeprecationWarning,
                stacklevel=warning_stacklevel,
            )

    return normalized_tags


def _normalize_value(value: Any, key: str) -> str:
    error = None
    try:
        serialized_value = seven.json.dumps(value)
    except TypeError:
        error = 'Could not JSON encode value "{value}"'
    if not error and not seven.json.loads(serialized_value) == value:
        error = f'JSON encoding "{serialized_value}" of value "{value}" is not equivalent to original value'
    if error:
        raise DagsterInvalidDefinitionError(
            f'Invalid value for tag "{key}", {error}. Tag values must be strings '
            "or meet the constraint that json.loads(json.dumps(value)) == value."
        )
    return serialized_value


def is_private_system_tag_key(tag) -> bool:
    return tag.startswith(SYSTEM_TAG_PREFIX) and tag not in USER_EDITABLE_SYSTEM_TAGS


def is_valid_tag_key(key: str) -> bool:
    return bool(VALID_TAG_KEY_REGEX.match(key))


def is_valid_strict_tag_value(key: str) -> bool:
    return bool(VALID_STRICT_TAG_VALUE_REGEX.match(key))



from collections.abc import Iterable, Sequence
from typing import Any, Union

from dagster import _check as check
from dagster._annotations import beta
from dagster._core.definitions.asset_check_factories.utils import (
    DEADLINE_CRON_PARAM_KEY,
    DEFAULT_FRESHNESS_SEVERITY,
    DEFAULT_FRESHNESS_TIMEZONE,
    FRESH_UNTIL_METADATA_KEY,
    FRESHNESS_PARAMS_METADATA_KEY,
    LAST_UPDATED_TIMESTAMP_METADATA_KEY,
    LATEST_CRON_TICK_METADATA_KEY,
    TIMEZONE_PARAM_KEY,
    assets_to_keys,
    ensure_no_duplicate_assets,
    freshness_multi_asset_check,
    retrieve_last_update_record,
    retrieve_timestamp_from_record,
)
from dagster._core.definitions.asset_check_result import AssetCheckResult
from dagster._core.definitions.asset_check_spec import AssetCheckSeverity
from dagster._core.definitions.asset_checks import AssetChecksDefinition
from dagster._core.definitions.assets import AssetsDefinition, SourceAsset
from dagster._core.definitions.events import AssetKey, CoercibleToAssetKey
from dagster._core.definitions.metadata import (
    JsonMetadataValue,
    MetadataValue,
    TimestampMetadataValue,
)
from dagster._core.definitions.time_window_partitions import TimeWindowPartitionsDefinition
from dagster._core.execution.context.compute import AssetCheckExecutionContext
from dagster._time import datetime_from_timestamp, get_current_timestamp
from dagster._utils.schedules import (
    get_latest_completed_cron_tick,
    get_next_cron_tick,
    is_valid_cron_string,
)


@beta
def build_time_partition_freshness_checks(
    *,
    assets: Sequence[Union[SourceAsset, CoercibleToAssetKey, AssetsDefinition]],
    deadline_cron: str,
    timezone: str = DEFAULT_FRESHNESS_TIMEZONE,
    severity: AssetCheckSeverity = DEFAULT_FRESHNESS_SEVERITY,
) -> Sequence[AssetChecksDefinition]:
    r"""Construct an `AssetChecksDefinition` that checks the freshness of the provided assets.

    This check passes if the asset is considered "fresh" by the time that execution begins. We
    consider an asset to be "fresh" if there exists a record for the most recent partition, once
    the deadline has passed.

    `deadline_cron` is a cron schedule that defines the deadline for when we should expect the most
    recent partition to arrive by. Once a tick of the cron schedule has passed, this check will fail
    if the most recent partition has not been observed/materialized.

    Let's say I have a daily-partitioned asset which runs every day at 8:00 AM UTC, and takes around
    45 minutes to complete. To account for operational delays, I would expect the asset to be done
    materializing every day by 9:00 AM UTC. I would set the `deadline_cron` to "0 9 \* \* \*". This
    means that starting at 9:00 AM, this check will expect a record to exist for the previous day's
    partition. Note that if the check runs at 8:59 AM, the deadline has not yet passed, and we'll
    instead be checking for the most recently passed deadline, which is yesterday (meaning the
    partition representing the day before yesterday).

    The timestamp of an observation record is the timestamp indicated by the
    "dagster/last_updated_timestamp" metadata key. The timestamp of a materialization record is the
    timestamp at which that record was created.

    The check will fail at runtime if a non-time-window partitioned asset is passed in.

    The check result will contain the following metadata:
    "dagster/freshness_params": A dictionary containing the parameters used to construct the
    check.
    "dagster/last_updated_time": (Only present if the asset has been observed/materialized before)
    The time of the most recent update to the asset.
    "dagster/overdue_seconds": (Only present if asset is overdue) The number of seconds that the
    asset is overdue by.
    "dagster/overdue_deadline_timestamp": The timestamp that we are expecting the asset to have
    arrived by. This is the timestamp of the most recent tick of the cron schedule.

    Examples:
        .. code-block:: python

            from dagster import build_time_partition_freshness_checks, AssetKey
            # A daily partitioned asset that is expected to be updated every day within 45 minutes
            # of 9:00 AM UTC
            from .somewhere import my_daily_scheduled_assets_def

            checks_def = build_time_partition_freshness_checks(
                [my_daily_scheduled_assets_def],
                deadline_cron="0 9 * * *",
            )


    Args:
        assets (Sequence[Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset]): The assets to
            construct checks for. For each passed in asset, there will be a corresponding
            constructed `AssetChecksDefinition`.
        deadline_cron (str): The check will pass if the partition time window most recently
            completed by the time of the last cron tick has been observed/materialized.
        timezone (Optional[str]): The timezone to use when calculating freshness and deadline. If
            not provided, defaults to "UTC".

    Returns:
        Sequence[AssetChecksDefinition]: `AssetChecksDefinition` objects which execute freshness
            checks for the provided assets.
    """
    check.str_param(timezone, "timezone")
    check.opt_str_param(deadline_cron, "deadline_cron")
    check.invariant(
        deadline_cron is None or is_valid_cron_string(deadline_cron), "Invalid cron string."
    )
    check.inst_param(severity, "severity", AssetCheckSeverity)
    check.sequence_param(assets, "assets")
    ensure_no_duplicate_assets(assets)
    return [
        _build_freshness_multi_check(
            asset_keys=assets_to_keys(assets),
            deadline_cron=deadline_cron,
            timezone=timezone,
            severity=severity,
        )
    ]


def _build_freshness_multi_check(
    asset_keys: Sequence[AssetKey],
    deadline_cron: str,
    timezone: str,
    severity: AssetCheckSeverity,
) -> AssetChecksDefinition:
    params_metadata: dict[str, Any] = {
        TIMEZONE_PARAM_KEY: timezone,
        DEADLINE_CRON_PARAM_KEY: deadline_cron,
    }

    @freshness_multi_asset_check(
        params_metadata=JsonMetadataValue(params_metadata), asset_keys=asset_keys
    )
    def the_check(context: AssetCheckExecutionContext) -> Iterable[AssetCheckResult]:
        for check_key in context.selected_asset_check_keys:
            asset_key = check_key.asset_key
            current_timestamp = get_current_timestamp()

            partitions_def = check.inst(
                context.job_def.asset_layer.asset_graph.get(asset_key).partitions_def,
                TimeWindowPartitionsDefinition,
            )
            current_time_in_freshness_tz = datetime_from_timestamp(current_timestamp, tz=timezone)
            deadline = get_latest_completed_cron_tick(
                deadline_cron, current_time_in_freshness_tz, timezone
            )
            deadline_in_partitions_def_tz = datetime_from_timestamp(
                deadline.timestamp(), tz=partitions_def.timezone
            )
            last_completed_time_window = check.not_none(
                partitions_def.get_last_partition_window(current_time=deadline_in_partitions_def_tz)
            )
            expected_partition_key = partitions_def.get_partition_key_range_for_time_window(
                last_completed_time_window
            ).start
            latest_record = retrieve_last_update_record(
                instance=context.instance, asset_key=asset_key, partition_key=expected_partition_key
            )
            passed = latest_record is not None

            metadata: dict[str, MetadataValue] = {
                FRESHNESS_PARAMS_METADATA_KEY: JsonMetadataValue(params_metadata),
                LATEST_CRON_TICK_METADATA_KEY: TimestampMetadataValue(deadline.timestamp()),
            }

            # Allows us to distinguish between the case where the asset has never been
            # observed/materialized, and the case where this partition in particular is missing
            latest_record_any_partition = retrieve_last_update_record(
                instance=context.instance, asset_key=asset_key, partition_key=None
            )

            if not passed and latest_record_any_partition is not None:
                # If this asset has been updated at all before, provide the time at which that
                # happened as additional metadata.
                metadata[LAST_UPDATED_TIMESTAMP_METADATA_KEY] = TimestampMetadataValue(
                    check.not_none(retrieve_timestamp_from_record(latest_record_any_partition))
                )
            elif passed:
                metadata[FRESH_UNTIL_METADATA_KEY] = TimestampMetadataValue(
                    get_next_cron_tick(
                        deadline_cron, current_time_in_freshness_tz, timezone
                    ).timestamp()
                )

            yield AssetCheckResult(
                passed=passed,
                description=_construct_description(
                    partition_key=expected_partition_key,
                    passed=passed,
                    any_records_exist_for_asset=latest_record_any_partition is not None,
                ),
                severity=severity,
                asset_key=asset_key,
                metadata=metadata,
            )

    return the_check


def _construct_description(
    partition_key: str,
    passed: bool,
    any_records_exist_for_asset: bool,
) -> str:
    if passed:
        return f"Asset is currently fresh, since partition {partition_key} has been observed/materialized."
    elif not any_records_exist_for_asset:
        return f"The asset has never been observed/materialized. We currently expect partition {partition_key} to have arrived."
    return (
        f"Asset is overdue. We expected partition {partition_key} to have arrived, and it has not."
    )



from collections import Counter
from collections.abc import Iterable, Mapping, Sequence
from functools import cached_property
from inspect import Parameter
from typing import AbstractSet, Any, Callable, NamedTuple, Optional, cast  # noqa: UP035

import dagster._check as check
from dagster._config.config_schema import UserConfigSchema
from dagster._core.decorator_utils import get_function_params, get_valid_name_permutations
from dagster._core.definitions.asset_check_spec import AssetCheckSpec
from dagster._core.definitions.asset_dep import AssetDep
from dagster._core.definitions.asset_in import AssetIn
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_out import AssetOut
from dagster._core.definitions.asset_spec import (
    SYSTEM_METADATA_KEY_DAGSTER_TYPE,
    SYSTEM_METADATA_KEY_IO_MANAGER_KEY,
    AssetExecutionType,
    AssetSpec,
)
from dagster._core.definitions.assets import (
    ASSET_SUBSET_INPUT_PREFIX,
    AssetsDefinition,
    get_partition_mappings_from_deps,
    stringify_asset_key_to_input_name,
)
from dagster._core.definitions.backfill_policy import BackfillPolicy
from dagster._core.definitions.decorators.op_decorator import _Op
from dagster._core.definitions.input import In
from dagster._core.definitions.op_definition import OpDefinition
from dagster._core.definitions.output import Out
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.partition_mapping import PartitionMapping
from dagster._core.definitions.policy import RetryPolicy
from dagster._core.definitions.resource_annotation import get_resource_args
from dagster._core.definitions.resource_definition import ResourceDefinition
from dagster._core.definitions.utils import NoValueSentinel
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._core.storage.tags import COMPUTE_KIND_TAG
from dagster._core.types.dagster_type import (
    Any as DagsterAny,
    DagsterType,
    Nothing,
)


def get_function_params_without_context_or_config_or_resources(
    fn: Callable[..., Any],
) -> list[Parameter]:
    params = get_function_params(fn)
    is_context_provided = len(params) > 0 and params[0].name in get_valid_name_permutations(
        "context"
    )
    input_params = params[1:] if is_context_provided else params

    resource_arg_names = {arg.name for arg in get_resource_args(fn)}

    new_input_args = []
    for input_arg in input_params:
        if input_arg.name != "config" and input_arg.name not in resource_arg_names:
            new_input_args.append(input_arg)

    return new_input_args


def validate_can_coexist(asset_in: AssetIn, asset_dep: AssetDep) -> None:
    """Validates that the asset_in and asset_dep can coexist peacefully on the same asset key.
    If both asset_in and asset_dep are set on the same asset key, expect that _no_ properties
    are set on AssetIn except for the key itself.
    """
    if (
        asset_in.metadata
        or asset_in.key_prefix
        or asset_in.dagster_type != NoValueSentinel
        or asset_in.partition_mapping is not None
    ):
        raise DagsterInvalidDefinitionError(
            f"Asset key '{asset_dep.asset_key.to_user_string()}' is used as both an input (via AssetIn) and a dependency (via AssetDep). If an asset key is used as an input and also set as a dependency, the input should only define the relationship between the asset key and the input name, or optionally set the input_manager_key. Any other properties should either not be set, or should be set on the dependency."
        )


def build_and_validate_named_ins(
    fn: Callable[..., Any],
    asset_ins: Mapping[str, AssetIn],
    deps: Optional[Iterable[AssetDep]],
) -> Mapping[AssetKey, "NamedIn"]:
    """Creates a mapping from AssetKey to (name of input, In object)."""
    deps_by_key = {dep.asset_key: dep for dep in deps} if deps else {}
    ins_by_asset_key = {
        asset_in.key if asset_in.key else AssetKey.from_coercible(input_name): asset_in
        for input_name, asset_in in asset_ins.items()
    }
    shared_keys_between_ins_and_deps = set(ins_by_asset_key.keys()) & set(deps_by_key.keys())
    if shared_keys_between_ins_and_deps:
        for shared_key in shared_keys_between_ins_and_deps:
            validate_can_coexist(ins_by_asset_key[shared_key], deps_by_key[shared_key])

    deps = check.opt_iterable_param(deps, "deps", AssetDep)

    new_input_args = get_function_params_without_context_or_config_or_resources(fn)

    non_var_input_param_names = [
        param.name for param in new_input_args if param.kind == Parameter.POSITIONAL_OR_KEYWORD
    ]
    has_kwargs = any(param.kind == Parameter.VAR_KEYWORD for param in new_input_args)

    all_input_names = set(non_var_input_param_names) | asset_ins.keys()

    if not has_kwargs:
        for in_key, asset_in in asset_ins.items():
            if in_key not in non_var_input_param_names and (
                not isinstance(asset_in.dagster_type, DagsterType)
                or not asset_in.dagster_type.is_nothing
            ):
                raise DagsterInvalidDefinitionError(
                    f"Key '{in_key}' in provided ins dict does not correspond to any of the names "
                    "of the arguments to the decorated function"
                )

    named_ins_by_asset_key: dict[AssetKey, NamedIn] = {}
    for input_name in all_input_names:
        asset_key = None

        if input_name in asset_ins:
            asset_key = asset_ins[input_name].key
            metadata = asset_ins[input_name].metadata or {}
            key_prefix = asset_ins[input_name].key_prefix
            input_manager_key = asset_ins[input_name].input_manager_key
            dagster_type = asset_ins[input_name].dagster_type
        else:
            metadata = {}
            key_prefix = None
            input_manager_key = None
            dagster_type = NoValueSentinel

        asset_key = asset_key or AssetKey(list(filter(None, [*(key_prefix or []), input_name])))

        named_ins_by_asset_key[asset_key] = NamedIn(
            input_name.replace("-", "_"),
            In(metadata=metadata, input_manager_key=input_manager_key, dagster_type=dagster_type),
        )

    for dep in deps:
        if dep.asset_key not in named_ins_by_asset_key:
            named_ins_by_asset_key[dep.asset_key] = NamedIn(
                stringify_asset_key_to_input_name(dep.asset_key),
                In(cast(type, Nothing)),
            )

    return named_ins_by_asset_key


def build_named_outs(asset_outs: Mapping[str, AssetOut]) -> Mapping[AssetKey, "NamedOut"]:
    """Creates a mapping from AssetKey to (name of output, Out object)."""
    named_outs_by_asset_key: dict[AssetKey, NamedOut] = {}
    for output_name, asset_out in asset_outs.items():
        out = asset_out.to_out()
        asset_key = asset_out.key or AssetKey(
            list(filter(None, [*(asset_out.key_prefix or []), output_name]))
        )

        named_outs_by_asset_key[asset_key] = NamedOut(output_name.replace("-", "_"), out)

    return named_outs_by_asset_key


def build_subsettable_named_ins(
    asset_ins: Mapping[AssetKey, tuple[str, In]],
    asset_outs: Mapping[AssetKey, tuple[str, Out]],
    internal_upstream_deps: Iterable[AbstractSet[AssetKey]],
) -> Mapping[AssetKey, "NamedIn"]:
    """Creates a mapping from AssetKey to (name of input, In object) for any asset key that is not
    currently an input, but may become one if this asset is subset.

    For example, if a subsettable multi-asset produces both A and C, where C depends on both A and
    some other asset B, there are some situations where executing just A and C without B will result
    in these assets being generated by different steps within the same job. In this case, we need
    a separate input to represent the fact that C depends on A.
    """
    # set of asset keys which are upstream of another asset, and are not currently inputs
    potential_deps = set().union(*internal_upstream_deps).difference(set(asset_ins.keys()))
    return {
        key: NamedIn(f"{ASSET_SUBSET_INPUT_PREFIX}{name}", In(Nothing))
        for key, (name, _) in asset_outs.items()
        if key in potential_deps
    }


class NamedIn(NamedTuple):
    input_name: str
    input: In


class NamedOut(NamedTuple):
    output_name: str
    output: Out


def make_keys_by_output_name(
    asset_outs: Mapping[AssetKey, tuple[str, Out]],
) -> Mapping[str, AssetKey]:
    return {output_name: asset_key for asset_key, (output_name, _) in asset_outs.items()}


def compute_required_resource_keys(
    required_resource_keys: AbstractSet[str],
    resource_defs: Mapping[str, ResourceDefinition],
    fn: Callable[..., Any],
    decorator_name: str,
) -> AbstractSet[str]:
    bare_required_resource_keys = set(required_resource_keys)
    resource_defs_keys = set(resource_defs.keys())
    required_resource_keys = bare_required_resource_keys | resource_defs_keys
    arg_resource_keys = {arg.name for arg in get_resource_args(fn)}
    check.param_invariant(
        len(bare_required_resource_keys or []) == 0 or len(arg_resource_keys) == 0,
        f"Cannot specify resource requirements in both {decorator_name} decorator and as"
        " arguments to the decorated function",
    )
    return required_resource_keys - arg_resource_keys


class DecoratorAssetsDefinitionBuilderArgs(NamedTuple):
    asset_deps: Mapping[str, set[AssetKey]]
    asset_in_map: Mapping[str, AssetIn]
    asset_out_map: Mapping[str, AssetOut]
    assets_def_resource_defs: Mapping[str, ResourceDefinition]
    backfill_policy: Optional[BackfillPolicy]
    can_subset: bool
    check_specs_by_output_name: Mapping[str, AssetCheckSpec]
    code_version: Optional[str]
    compute_kind: Optional[str]
    config_schema: Optional[UserConfigSchema]
    decorator_name: str
    group_name: Optional[str]
    name: Optional[str]
    op_def_resource_defs: Mapping[str, ResourceDefinition]
    op_description: Optional[str]
    op_tags: Optional[Mapping[str, Any]]
    partitions_def: Optional[PartitionsDefinition]
    required_resource_keys: AbstractSet[str]
    retry_policy: Optional[RetryPolicy]
    retry_policy: Optional[RetryPolicy]
    specs: Sequence[AssetSpec]
    upstream_asset_deps: Optional[Iterable[AssetDep]]
    execution_type: Optional[AssetExecutionType]
    pool: Optional[str]

    @property
    def check_specs(self) -> Sequence[AssetCheckSpec]:
        return list(self.check_specs_by_output_name.values())


class DecoratorAssetsDefinitionBuilder:
    def __init__(
        self,
        *,
        named_ins_by_asset_key: Mapping[AssetKey, NamedIn],
        named_outs_by_asset_key: Mapping[AssetKey, NamedOut],
        internal_deps: Mapping[AssetKey, set[AssetKey]],
        op_name: str,
        args: DecoratorAssetsDefinitionBuilderArgs,
        fn: Callable[..., Any],
    ) -> None:
        self.named_outs_by_asset_key = named_outs_by_asset_key
        self.internal_deps = internal_deps
        self.op_name = op_name
        self.args = args
        self.fn = fn

        self.named_ins_by_asset_key = (
            (
                {
                    **named_ins_by_asset_key,
                    **build_subsettable_named_ins(
                        named_ins_by_asset_key,
                        named_outs_by_asset_key,
                        self.internal_deps.values(),
                    ),
                }
            )
            if self.args.can_subset and self.internal_deps
            else named_ins_by_asset_key
        )

    @staticmethod
    def for_multi_asset(
        *, fn: Callable[..., Any], args: DecoratorAssetsDefinitionBuilderArgs
    ) -> "DecoratorAssetsDefinitionBuilder":
        op_name = args.name or fn.__name__

        if args.asset_out_map and args.specs:
            raise DagsterInvalidDefinitionError("Must specify only outs or specs but not both.")

        if args.compute_kind and args.specs and any(spec.kinds for spec in args.specs):
            raise DagsterInvalidDefinitionError(
                "Can not specify compute_kind on both the @multi_asset and kinds on AssetSpecs."
            )

        if args.specs:
            check.invariant(
                args.decorator_name == "@multi_asset", "Only hit this code path in multi_asset."
            )
            if args.upstream_asset_deps:
                raise DagsterInvalidDefinitionError(
                    "Can not pass deps and specs to @multi_asset, specify deps on the AssetSpecs"
                    " directly."
                )
            if args.asset_deps:
                raise DagsterInvalidDefinitionError(
                    "Can not pass internal_asset_deps and specs to @multi_asset, specify deps on"
                    " the AssetSpecs directly."
                )
            return DecoratorAssetsDefinitionBuilder.from_multi_asset_specs(
                fn=fn,
                op_name=op_name,
                passed_args=args,
                asset_specs=args.specs,
                can_subset=args.can_subset,
                asset_in_map=args.asset_in_map,
            )

        return DecoratorAssetsDefinitionBuilder.from_asset_outs_in_asset_centric_decorator(
            fn=fn,
            op_name=op_name,
            asset_in_map=args.asset_in_map,
            asset_out_map=args.asset_out_map,
            asset_deps=args.asset_deps,
            upstream_asset_deps=args.upstream_asset_deps,
            passed_args=args,
        )

    @staticmethod
    def from_multi_asset_specs(
        *,
        fn: Callable[..., Any],
        op_name: str,
        asset_specs: Sequence[AssetSpec],
        can_subset: bool,
        asset_in_map: Mapping[str, AssetIn],
        passed_args: DecoratorAssetsDefinitionBuilderArgs,
    ) -> "DecoratorAssetsDefinitionBuilder":
        check.param_invariant(passed_args.specs, "passed_args", "Must use specs in this codepath")

        named_outs_by_asset_key: Mapping[AssetKey, NamedOut] = {}
        for asset_spec in asset_specs:
            output_name = asset_spec.key.to_python_identifier()
            if SYSTEM_METADATA_KEY_DAGSTER_TYPE in asset_spec.metadata:
                dagster_type = asset_spec.metadata[SYSTEM_METADATA_KEY_DAGSTER_TYPE]
            elif asset_spec.metadata.get(SYSTEM_METADATA_KEY_IO_MANAGER_KEY):
                dagster_type = DagsterAny
            else:
                dagster_type = Nothing
            named_outs_by_asset_key[asset_spec.key] = NamedOut(
                output_name,
                Out(
                    dagster_type=dagster_type,
                    is_required=not (can_subset or asset_spec.skippable),
                    description=asset_spec.description,
                    code_version=asset_spec.code_version,
                    metadata=asset_spec.metadata,
                    io_manager_key=asset_spec.metadata.get(SYSTEM_METADATA_KEY_IO_MANAGER_KEY),
                ),
            )

        upstream_deps = {}
        for spec in asset_specs:
            for dep in spec.deps:
                if dep.asset_key not in named_outs_by_asset_key:
                    upstream_deps[dep.asset_key] = dep
                if dep.asset_key in named_outs_by_asset_key and dep.partition_mapping is not None:
                    # self-dependent asset also needs to be considered an upstream_key
                    upstream_deps[dep.asset_key] = dep

        # get which asset keys have inputs set
        named_ins_by_asset_key = build_and_validate_named_ins(
            fn, asset_in_map, deps=upstream_deps.values()
        )
        # We expect that asset_ins are a subset of asset_deps. The reason we do not check this in
        # `build_and_validate_named_ins` is because in other decorator pathways, we allow for argument-based
        # dependencies which are not specified in deps (such as the asset decorator).
        validate_named_ins_subset_of_deps(named_ins_by_asset_key, upstream_deps)

        internal_deps = {
            spec.key: {dep.asset_key for dep in spec.deps}
            for spec in asset_specs
            if spec.deps is not None
        }

        _validate_check_specs_target_relevant_asset_keys(
            passed_args.check_specs, [spec.key for spec in asset_specs]
        )

        return DecoratorAssetsDefinitionBuilder(
            named_ins_by_asset_key=named_ins_by_asset_key,
            named_outs_by_asset_key=named_outs_by_asset_key,
            internal_deps=internal_deps,
            op_name=op_name,
            args=passed_args,
            fn=fn,
        )

    @staticmethod
    def from_asset_outs_in_asset_centric_decorator(
        *,
        fn: Callable[..., Any],
        op_name: str,
        asset_in_map: Mapping[str, AssetIn],
        asset_out_map: Mapping[str, AssetOut],
        asset_deps: Mapping[str, set[AssetKey]],
        upstream_asset_deps: Optional[Iterable[AssetDep]],
        passed_args: DecoratorAssetsDefinitionBuilderArgs,
    ):
        check.param_invariant(
            not passed_args.specs, "args", "This codepath for non-spec based create"
        )
        named_ins_by_asset_key = build_and_validate_named_ins(
            fn,
            asset_in_map,
            deps=upstream_asset_deps or set(),
        )
        named_outs_by_asset_key = build_named_outs(asset_out_map)

        # validate that the asset_ins are a subset of the upstream asset_deps.
        upstream_internal_asset_keys = set().union(*asset_deps.values())
        asset_in_keys = set(named_ins_by_asset_key.keys())
        if asset_deps and not asset_in_keys.issubset(upstream_internal_asset_keys):
            invalid_asset_in_keys = asset_in_keys - upstream_internal_asset_keys
            check.failed(
                f"Invalid asset dependencies: `{invalid_asset_in_keys}` specified as asset"
                " inputs, but are not specified in `internal_asset_deps`. Asset inputs must"
                " be associated with an output produced by the asset."
            )

        # validate that the asset_deps make sense
        valid_asset_deps = asset_in_keys | set(named_outs_by_asset_key.keys())
        for out_name, asset_keys in asset_deps.items():
            if asset_out_map and out_name not in asset_out_map:
                check.failed(
                    f"Invalid out key '{out_name}' supplied to `internal_asset_deps` argument"
                    f" for multi-asset {op_name}. Must be one of the outs for this multi-asset"
                    f" {list(asset_out_map.keys())[:20]}.",
                )
            invalid_asset_deps = asset_keys.difference(valid_asset_deps)
            check.invariant(
                not invalid_asset_deps,
                f"Invalid asset dependencies: {invalid_asset_deps} specified in"
                f" `internal_asset_deps` argument for multi-asset '{op_name}' on key"
                f" '{out_name}'. Each specified asset key must be associated with an input to"
                " the asset or produced by this asset. Valid keys:"
                f" {list(valid_asset_deps)[:20]}",
            )

        keys_by_output_name = make_keys_by_output_name(named_outs_by_asset_key)
        internal_deps = {keys_by_output_name[name]: asset_deps[name] for name in asset_deps}

        _validate_check_specs_target_relevant_asset_keys(
            passed_args.check_specs, list(named_outs_by_asset_key.keys())
        )

        return DecoratorAssetsDefinitionBuilder(
            named_ins_by_asset_key=named_ins_by_asset_key,
            named_outs_by_asset_key=named_outs_by_asset_key,
            internal_deps=internal_deps,
            op_name=op_name,
            args=passed_args,
            fn=fn,
        )

    @property
    def group_name(self) -> Optional[str]:
        return self.args.group_name

    @cached_property
    def outs_by_output_name(self) -> Mapping[str, Out]:
        return dict(self.named_outs_by_asset_key.values())

    @cached_property
    def asset_keys_by_input_name(self) -> Mapping[str, AssetKey]:
        return {
            in_mapping.input_name: asset_key
            for asset_key, in_mapping in self.named_ins_by_asset_key.items()
        }

    @cached_property
    def asset_keys_by_output_name(self) -> Mapping[str, AssetKey]:
        return {
            out_mapping.output_name: asset_key
            for asset_key, out_mapping in self.named_outs_by_asset_key.items()
        }

    @cached_property
    def asset_keys(self) -> set[AssetKey]:
        return set(self.named_outs_by_asset_key.keys())

    @cached_property
    def check_specs_by_output_name(self) -> Mapping[str, AssetCheckSpec]:
        return self.args.check_specs_by_output_name

    @cached_property
    def check_outs_by_output_name(self) -> Mapping[str, Out]:
        return {
            output_name: Out(dagster_type=None, is_required=not self.args.can_subset)
            for output_name in self.check_specs_by_output_name.keys()
        }

    @cached_property
    def combined_outs_by_output_name(self) -> Mapping[str, Out]:
        return {
            **self.outs_by_output_name,
            **self.check_outs_by_output_name,
        }

    @cached_property
    def overlapping_output_names(self) -> set[str]:
        return set(self.outs_by_output_name.keys()) & set(self.check_outs_by_output_name.keys())

    @cached_property
    def ins_by_input_names(self) -> Mapping[str, In]:
        return {in_name: in_obj for in_name, in_obj in self.named_ins_by_asset_key.values()}

    @cached_property
    def asset_keys_by_input_names(self) -> Mapping[str, AssetKey]:
        return {
            in_mapping.input_name: asset_key
            for asset_key, in_mapping in self.named_ins_by_asset_key.items()
        }

    @cached_property
    def partition_mappings(self) -> Mapping[AssetKey, PartitionMapping]:
        partition_mappings = {
            self.asset_keys_by_input_names[input_name]: asset_in.partition_mapping
            for input_name, asset_in in self.args.asset_in_map.items()
            if asset_in.partition_mapping is not None
        }

        if not self.args.upstream_asset_deps:
            return partition_mappings

        return get_partition_mappings_from_deps(
            partition_mappings=partition_mappings,
            deps=self.args.upstream_asset_deps,
            asset_name=self.op_name,
        )

    @cached_property
    def required_resource_keys(self) -> AbstractSet[str]:
        return compute_required_resource_keys(
            required_resource_keys=self.args.required_resource_keys,
            resource_defs=self.args.op_def_resource_defs,
            fn=self.fn,
            decorator_name=self.args.decorator_name,
        )

    def create_op_definition(self) -> OpDefinition:
        return _Op(
            name=self.op_name,
            description=self.args.op_description,
            ins=self.ins_by_input_names,
            out=self.combined_outs_by_output_name,
            required_resource_keys=self.required_resource_keys,
            tags={
                **({COMPUTE_KIND_TAG: self.args.compute_kind} if self.args.compute_kind else {}),
                **(self.args.op_tags or {}),
            },
            config_schema=self.args.config_schema,
            retry_policy=self.args.retry_policy,
            code_version=self.args.code_version,
            pool=self.args.pool,
        )(self.fn)

    def create_assets_definition(self) -> AssetsDefinition:
        return AssetsDefinition.dagster_internal_init(
            keys_by_input_name=self.asset_keys_by_input_names,
            keys_by_output_name=self.asset_keys_by_output_name,
            node_def=self.create_op_definition(),
            can_subset=self.args.can_subset,
            resource_defs=self.args.assets_def_resource_defs,
            backfill_policy=self.args.backfill_policy,
            check_specs_by_output_name=self.check_specs_by_output_name,
            specs=self.specs,
            is_subset=False,
            selected_asset_keys=None,  # not a subset so this is None
            selected_asset_check_keys=None,  # not a subset so this is none
            execution_type=self.args.execution_type,
        )

    @cached_property
    def specs(self) -> Sequence[AssetSpec]:
        if self.args.specs:
            specs = self.args.specs
            self._validate_spec_partitions_defs(specs, self.args.partitions_def)
        else:
            specs = self._synthesize_specs()

        check.invariant(
            not self.group_name
            or all(
                (spec.group_name is None or spec.group_name == self.group_name) for spec in specs
            ),
            "Cannot set group_name parameter on multi_asset if one or more of the"
            " AssetSpecs/AssetOuts supplied to this multi_asset have a group_name defined.",
        )

        if not self.group_name and not self.args.partitions_def:
            return specs

        return [
            spec.replace_attributes(
                group_name=spec.group_name or self.group_name,
                partitions_def=spec.partitions_def or self.args.partitions_def,
            )
            for spec in specs
        ]

    def _validate_spec_partitions_defs(
        self, specs: Sequence[AssetSpec], partitions_def: Optional[PartitionsDefinition]
    ) -> Optional[PartitionsDefinition]:
        any_spec_has_partitions_def = False
        any_spec_has_no_partitions_def = False
        if partitions_def is not None:
            for spec in specs:
                if spec.partitions_def is not None and spec.partitions_def != partitions_def:
                    check.failed(
                        f"AssetSpec for {spec.key.to_user_string()} has partitions_def "
                        f"(type={type(spec.partitions_def)}) which is different than the "
                        f"partitions_def provided to AssetsDefinition (type={type(partitions_def)}).",
                    )

                any_spec_has_partitions_def = (
                    any_spec_has_partitions_def or spec.partitions_def is not None
                )
                any_spec_has_no_partitions_def = (
                    any_spec_has_no_partitions_def or spec.partitions_def is None
                )

        if (
            partitions_def is not None
            and any_spec_has_partitions_def
            and any_spec_has_no_partitions_def
        ):
            check.failed(
                "If partitions_def is provided, then either all specs must have that PartitionsDefinition or none."
            )

    def _synthesize_specs(self) -> Sequence[AssetSpec]:
        resolved_specs = []
        input_deps_by_key = {
            key: AssetDep(asset=key, partition_mapping=self.partition_mappings.get(key))
            for key in self.asset_keys_by_input_names.values()
        }
        input_deps = list(input_deps_by_key.values())
        for output_name, asset_out in self.args.asset_out_map.items():
            key = self.asset_keys_by_output_name[output_name]
            if self.args.asset_deps:
                deps = [
                    input_deps_by_key.get(
                        dep_key,
                        AssetDep(
                            asset=dep_key,
                            partition_mapping=self.partition_mappings.get(key),
                        ),
                    )
                    for dep_key in self.args.asset_deps.get(output_name, [])
                ]
            else:
                deps = input_deps

            resolved_specs.append(
                asset_out.to_spec(key, deps=deps, partitions_def=self.args.partitions_def)
            )

        specs = resolved_specs
        return specs


def validate_and_assign_output_names_to_check_specs(
    check_specs: Optional[Sequence[AssetCheckSpec]], valid_asset_keys: Sequence[AssetKey]
) -> Mapping[str, AssetCheckSpec]:
    _validate_check_specs_target_relevant_asset_keys(check_specs, valid_asset_keys)
    return create_check_specs_by_output_name(check_specs)


def create_check_specs_by_output_name(
    check_specs: Optional[Sequence[AssetCheckSpec]],
) -> Mapping[str, AssetCheckSpec]:
    checks_by_output_name = {
        spec.get_python_identifier(): spec
        for spec in check.opt_sequence_param(check_specs, "check_specs", of_type=AssetCheckSpec)
    }
    if check_specs and len(checks_by_output_name) != len(check_specs):
        duplicates = {
            item: count
            for item, count in Counter(
                [(spec.asset_key, spec.name) for spec in check_specs]
            ).items()
            if count > 1
        }

        raise DagsterInvalidDefinitionError(f"Duplicate check specs: {duplicates}")

    return checks_by_output_name


def _validate_check_specs_target_relevant_asset_keys(
    check_specs: Optional[Sequence[AssetCheckSpec]], valid_asset_keys: Sequence[AssetKey]
) -> None:
    for spec in check_specs or []:
        if spec.asset_key not in valid_asset_keys:
            raise DagsterInvalidDefinitionError(
                f"Invalid asset key {spec.asset_key} in check spec {spec.name}. Must be one of"
                f" {valid_asset_keys}"
            )


def validate_named_ins_subset_of_deps(
    named_ins_per_key: Mapping[AssetKey, NamedIn],
    asset_deps_by_key: Mapping[AssetKey, AssetDep],
) -> None:
    """Validates that the asset_ins are a subset of the asset_deps. This is a common validation
    that we need to do in multiple places, so we've factored it out into a helper function.
    """
    asset_dep_keys = set(asset_deps_by_key.keys())
    asset_in_keys = set(named_ins_per_key.keys())

    if asset_in_keys - asset_dep_keys:
        invalid_asset_in_keys = asset_in_keys - asset_dep_keys
        raise DagsterInvalidDefinitionError(
            f"Invalid asset dependencies: `{invalid_asset_in_keys}` specified as AssetIns, but"
            " are not specified as `AssetDep` objects on any constituent AssetSpec objects. Asset inputs must be associated with an"
            " output produced by the asset."
        )



from collections.abc import Mapping
from enum import Enum
from functools import cached_property
from typing import TYPE_CHECKING, Annotated, Optional

from dagster._record import ImportFrom, record
from dagster._utils.error import SerializableErrorInfo

if TYPE_CHECKING:
    from dagster._core.definitions.remote_asset_graph import RemoteWorkspaceAssetGraph
    from dagster._core.remote_representation import CodeLocation, CodeLocationOrigin


# For locations that are loaded asynchronously
class CodeLocationLoadStatus(Enum):
    LOADING = "LOADING"  # Waiting for location to load or update
    LOADED = "LOADED"  # Finished loading (may be an error)


@record
class CodeLocationEntry:
    origin: Annotated["CodeLocationOrigin", ImportFrom("dagster._core.remote_representation")]
    code_location: Optional[
        Annotated["CodeLocation", ImportFrom("dagster._core.remote_representation")]
    ]
    load_error: Optional[SerializableErrorInfo]
    load_status: CodeLocationLoadStatus
    display_metadata: Mapping[str, str]
    update_timestamp: float
    version_key: str


@record
class CodeLocationStatusEntry:
    """Slimmer version of WorkspaceLocationEntry, containing the minimum set of information required
    to know whether the workspace needs reloading.
    """

    location_name: str
    load_status: CodeLocationLoadStatus
    update_timestamp: float
    version_key: str


@record
class CurrentWorkspace:
    code_location_entries: Mapping[str, CodeLocationEntry]

    @cached_property
    def asset_graph(self) -> "RemoteWorkspaceAssetGraph":
        from dagster._core.definitions.remote_asset_graph import RemoteWorkspaceAssetGraph

        return RemoteWorkspaceAssetGraph.build(self)

    def with_code_location(self, name: str, entry: CodeLocationEntry) -> "CurrentWorkspace":
        return CurrentWorkspace(code_location_entries={**self.code_location_entries, name: entry})


def location_status_from_location_entry(
    entry: CodeLocationEntry,
) -> CodeLocationStatusEntry:
    return CodeLocationStatusEntry(
        location_name=entry.origin.location_name,
        load_status=entry.load_status,
        update_timestamp=entry.update_timestamp,
        version_key=entry.version_key,
    )



import os
import shutil
import tempfile
from collections import defaultdict
from collections.abc import Mapping
from contextlib import contextmanager
from typing import Any, Dict, List, Optional, Set, Type, Union, cast  # noqa: F401, UP035

# top-level include is dangerous in terms of incurring circular deps
from dagster import (
    DagsterInvariantViolationError,
    DependencyDefinition,
    NodeInvocation,
    _check as check,
)
from dagster._config import Field, StringSource
from dagster._config.config_schema import UserConfigSchema
from dagster._core.definitions import (
    GraphDefinition,
    InputMapping,
    JobDefinition,
    OpDefinition,
    OutputMapping,
)
from dagster._core.definitions.dependency import Node
from dagster._core.definitions.executor_definition import in_process_executor
from dagster._core.definitions.job_base import InMemoryJob
from dagster._core.definitions.logger_definition import LoggerDefinition
from dagster._core.definitions.resource_definition import ScopedResourcesBuilder
from dagster._core.events import DagsterEventType
from dagster._core.execution.api import create_execution_plan
from dagster._core.execution.context.system import PlanExecutionContext
from dagster._core.execution.context_creation_job import (
    create_context_creation_data,
    create_execution_data,
    create_executor,
    create_log_manager,
    create_plan_data,
)
from dagster._core.execution.execute_in_process_result import ExecuteInProcessResult
from dagster._core.instance import DagsterInstance
from dagster._core.scheduler import Scheduler
from dagster._core.storage.dagster_run import DagsterRun
from dagster._core.storage.event_log.base import EventLogConnection
from dagster._core.storage.event_log.sqlite.sqlite_event_log import SqliteEventLogStorage
from dagster._core.storage.sqlite_storage import SqliteStorageConfig
from dagster._core.utility_ops import create_stub_op
from dagster._serdes import ConfigurableClass
from dagster._serdes.config_class import ConfigurableClassData
from dagster._utils.concurrency import ConcurrencyClaimStatus

# re-export
from dagster._utils.temp_file import (
    get_temp_dir as get_temp_dir,
    get_temp_file_handle as get_temp_file_handle,
    get_temp_file_handle_with_data as get_temp_file_handle_with_data,
    get_temp_file_name as get_temp_file_name,
    get_temp_file_name_with_data as get_temp_file_name_with_data,
    get_temp_file_names as get_temp_file_names,
)


def create_test_pipeline_execution_context(
    logger_defs: Optional[Mapping[str, LoggerDefinition]] = None,
) -> PlanExecutionContext:
    loggers = check.opt_mapping_param(
        logger_defs, "logger_defs", key_type=str, value_type=LoggerDefinition
    )
    job_def = GraphDefinition(
        name="test_legacy_context",
        node_defs=[],
    ).to_job(executor_def=in_process_executor, logger_defs=logger_defs)
    run_config: dict[str, dict[str, dict]] = {"loggers": {key: {} for key in loggers}}
    dagster_run = DagsterRun(job_name="test_legacy_context", run_config=run_config)
    instance = DagsterInstance.ephemeral()
    execution_plan = create_execution_plan(job=job_def, run_config=run_config)
    creation_data = create_context_creation_data(
        InMemoryJob(job_def),
        execution_plan,
        run_config,
        dagster_run,
        instance,
    )
    log_manager = create_log_manager(creation_data)
    scoped_resources_builder = ScopedResourcesBuilder()
    executor = create_executor(creation_data)

    return PlanExecutionContext(
        plan_data=create_plan_data(creation_data, True, executor.retries),
        execution_data=create_execution_data(
            context_creation_data=creation_data,
            scoped_resources_builder=scoped_resources_builder,
        ),
        log_manager=log_manager,
        output_capture=None,
    )


def _dep_key_of(node: Node) -> NodeInvocation:
    return NodeInvocation(node.definition.name, node.name)


def build_job_with_input_stubs(
    job_def: JobDefinition, inputs: Mapping[str, Mapping[str, object]]
) -> JobDefinition:
    check.inst_param(job_def, "pipeline_def", JobDefinition)
    check.mapping_param(inputs, "inputs", key_type=str, value_type=dict)

    deps: dict[NodeInvocation, dict[str, object]] = defaultdict(dict)
    for node_name, dep_dict in job_def.dependencies.items():
        for input_name, dep in dep_dict.items():
            deps[node_name][input_name] = dep

    stub_node_defs = []

    for node_name, input_dict in inputs.items():
        if not job_def.has_node_named(node_name):
            raise DagsterInvariantViolationError(
                f"You are injecting an input value for node {node_name} "
                f"into pipeline {job_def.name} but that node was not found"
            )

        node = job_def.get_node_named(node_name)
        for input_name, input_value in input_dict.items():
            stub_node_def = create_stub_op(
                f"__stub_{node_name}_{input_name}",
                input_value,
            )
            stub_node_defs.append(stub_node_def)
            deps[_dep_key_of(node)][input_name] = DependencyDefinition(stub_node_def.name)

    return JobDefinition(
        name=job_def.name + "_stubbed",
        graph_def=GraphDefinition(
            node_defs=[*job_def.top_level_node_defs, *stub_node_defs],
            dependencies=deps,  # type: ignore
        ),
        resource_defs=job_def.resource_defs,
    )


def wrap_op_in_graph(
    op_def: OpDefinition,
    tags: Optional[Mapping[str, Any]] = None,
    do_input_mapping: bool = True,
    do_output_mapping: bool = True,
) -> GraphDefinition:
    """Wraps op in a graph with the same inputs/outputs as the original op."""
    check.opt_mapping_param(tags, "tags", key_type=str)

    if do_input_mapping:
        input_mappings = []
        for input_name in op_def.ins.keys():
            # create an input mapping to the inner node with the same name.
            input_mappings.append(
                InputMapping(
                    graph_input_name=input_name,
                    mapped_node_name=op_def.name,
                    mapped_node_input_name=input_name,
                )
            )
    else:
        input_mappings = None

    if do_output_mapping:
        output_mappings = []
        for output_name in op_def.outs.keys():
            out = op_def.outs[output_name]
            output_mappings.append(
                OutputMapping(
                    graph_output_name=output_name,
                    mapped_node_name=op_def.name,
                    mapped_node_output_name=output_name,
                    from_dynamic_mapping=out.is_dynamic,
                )
            )
    else:
        output_mappings = None

    return GraphDefinition(
        name=f"wraps_{op_def.name}",
        node_defs=[op_def],
        input_mappings=input_mappings,
        output_mappings=output_mappings,
        tags=tags,
    )


def wrap_op_in_graph_and_execute(
    op_def: OpDefinition,
    resources: Optional[Mapping[str, Any]] = None,
    input_values: Optional[Mapping[str, Any]] = None,
    tags: Optional[Mapping[str, Any]] = None,
    run_config: Optional[Mapping[str, object]] = None,
    raise_on_error: bool = True,
    do_input_mapping: bool = True,
    do_output_mapping: bool = True,
    logger_defs: Optional[Mapping[str, LoggerDefinition]] = None,
) -> ExecuteInProcessResult:
    """Execute a single op in an ephemeral, in-process job.

    Intended to support unit tests. Input values may be passed directly, and no job need be
    specified -- an ephemeral one will be constructed.

    Args:
        op_def (OpDefinition): The op to execute.
        resources (Mapping[str, Any]): Resources that will be passed to `execute_in_process`.
        input_values (Optional[Dict[str, Any]]): A dict of input names to input values, used to
            pass inputs to the solid directly. You may also use the ``run_config`` to
            configure any inputs that are configurable.
        tags (Optional[Dict[str, Any]]): Arbitrary key-value pairs that will be added to pipeline
            logs.
        run_config (Optional[dict]): The configuration that parameterized this
            execution, as a dict.
        raise_on_error (Optional[bool]): Whether or not to raise exceptions when they occur.
            Defaults to ``True``, since this is the most useful behavior in test.
        do_input_mapping (bool): Whether to map the op inputs to the outputs of the graph
            constructed around it.
        do_output_mapping (bool): Whether to map the op outputs to the outputs of the graph
            constructed around it.

    Returns:
        Union[CompositeSolidExecutionResult, SolidExecutionResult]: The result of executing the
        solid.
    """
    return (
        wrap_op_in_graph(
            op_def, tags, do_input_mapping=do_input_mapping, do_output_mapping=do_output_mapping
        )
        .to_job(logger_defs=logger_defs)
        .execute_in_process(
            resources=resources,
            input_values=input_values,
            raise_on_error=raise_on_error,
            run_config=run_config,
        )
    )


@contextmanager
def copy_directory(src):
    with tempfile.TemporaryDirectory() as temp_dir:
        dst = os.path.join(temp_dir, os.path.basename(src))
        shutil.copytree(src, dst)
        yield dst


class FilesystemTestScheduler(Scheduler, ConfigurableClass):
    """This class is used in dagster core and dagster_graphql to test the scheduler's interactions
    with schedule storage, which are implemented in the methods defined on the base Scheduler class.
    Therefore, the following methods used to actually schedule jobs (e.g. create and remove cron jobs
    on a cron tab) are left unimplemented.
    """

    def __init__(self, artifacts_dir: str, inst_data: object = None):
        check.str_param(artifacts_dir, "artifacts_dir")
        self._artifacts_dir = artifacts_dir
        self._inst_data = inst_data

    @property
    def inst_data(self) -> object:
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {"base_dir": str}

    @classmethod
    def from_config_value(
        cls, inst_data: object, config_value: Mapping[str, object]
    ) -> "FilesystemTestScheduler":
        artifacts_dir = cast(str, config_value["base_dir"])
        return FilesystemTestScheduler(artifacts_dir=artifacts_dir, inst_data=inst_data)

    def debug_info(self) -> str:
        return ""

    def get_logs_path(self, _instance: DagsterInstance, schedule_origin_id: str) -> str:
        check.str_param(schedule_origin_id, "schedule_origin_id")
        return os.path.join(self._artifacts_dir, "logs", schedule_origin_id, "scheduler.log")

    def wipe(self, instance: DagsterInstance) -> None:
        pass


class TestStorageConfig(SqliteStorageConfig):
    # interval to sleep between claim checks
    sleep_interval: int


class ConcurrencyEnabledSqliteTestEventLogStorage(SqliteEventLogStorage, ConfigurableClass):
    """Sqlite is sorta supported for concurrency, as long as the rate of concurrent writes is tolerably
    low.  Officially, we should not support, but in the spirit of getting code coverage in the core
    dagster package, let's mark it as that.
    """

    __test__ = False

    def __init__(
        self,
        base_dir: str,
        sleep_interval: Optional[float] = None,
        inst_data: Optional[ConfigurableClassData] = None,
    ):
        self._sleep_interval = sleep_interval
        self._check_calls = defaultdict(int)
        self._records_for_run_calls = defaultdict(int)
        super().__init__(base_dir, inst_data)

    @classmethod
    def config_type(cls) -> UserConfigSchema:
        return {"base_dir": StringSource, "sleep_interval": Field(float, is_required=False)}

    @classmethod
    def from_config_value(
        cls, inst_data: Optional[ConfigurableClassData], config_value: TestStorageConfig
    ) -> "ConcurrencyEnabledSqliteTestEventLogStorage":
        return ConcurrencyEnabledSqliteTestEventLogStorage(inst_data=inst_data, **config_value)

    @property
    def supports_global_concurrency_limits(self) -> bool:
        return True

    def get_records_for_run(
        self,
        run_id: str,
        cursor: Optional[str] = None,
        of_type: Optional[Union[DagsterEventType, set[DagsterEventType]]] = None,
        limit: Optional[int] = None,
        ascending: bool = True,
    ) -> EventLogConnection:
        self._records_for_run_calls[run_id] = self._records_for_run_calls[run_id] + 1
        return super().get_records_for_run(run_id, cursor, of_type, limit, ascending)

    def get_check_calls(self, step_key: str) -> int:
        return self._check_calls[step_key]

    def get_records_for_run_calls(self, run_id: str) -> int:
        return self._records_for_run_calls[run_id]

    def check_concurrency_claim(
        self, concurrency_key: str, run_id: str, step_key: str
    ) -> ConcurrencyClaimStatus:
        self._check_calls[step_key] += 1
        claim_status = super().check_concurrency_claim(concurrency_key, run_id, step_key)
        if not self._sleep_interval:
            return claim_status
        return claim_status.with_sleep_interval(float(self._sleep_interval))


def get_all_direct_subclasses_of_marker(marker_interface_cls: type) -> list[type]:
    import dagster as dagster

    return [
        symbol
        for symbol in dagster.__dict__.values()
        if isinstance(symbol, type)
        and issubclass(symbol, marker_interface_cls)
        and marker_interface_cls
        in symbol.__bases__  # ensure that the class is a direct subclass of marker_interface_cls (not a subclass of a subclass)
    ]



import os
from abc import ABC, abstractmethod
from collections.abc import Generator, Iterator, Sequence
from contextlib import contextmanager
from enum import Enum
from typing import IO, Callable, Final, NamedTuple, Optional

from typing_extensions import Self

import dagster._check as check
from dagster._core.captured_log_api import LogLineCursor
from dagster._core.instance import MayHaveInstanceWeakref, T_DagsterInstance
from dagster._record import record
from dagster._serdes import whitelist_for_serdes

MAX_BYTES_CHUNK_READ: Final = 1048576  # 1 MB


class ComputeIOType(Enum):
    STDOUT = "stdout"
    STDERR = "stderr"


@whitelist_for_serdes
@record
class LogRetrievalShellCommand:
    stdout: Optional[str]
    stderr: Optional[str]


class CapturedLogContext(
    NamedTuple(
        "_CapturedLogContext",
        [
            ("log_key", Sequence[str]),
            ("external_url", Optional[str]),
            ("external_stdout_url", Optional[str]),
            ("external_stderr_url", Optional[str]),
            ("shell_cmd", Optional[LogRetrievalShellCommand]),
        ],
    )
):
    """Object representing the context in which logs are captured.  Can be used by external logging
    sidecar implementations to point the Dagster UI to an external url to view compute logs instead of a
    Dagster-managed location.
    """

    def __new__(
        cls,
        log_key: Sequence[str],
        external_stdout_url: Optional[str] = None,
        external_stderr_url: Optional[str] = None,
        external_url: Optional[str] = None,
        shell_cmd: Optional[LogRetrievalShellCommand] = None,
    ):
        if external_url and (external_stdout_url or external_stderr_url):
            check.failed(
                "Cannot specify both `external_url` and one of"
                " `external_stdout_url`/`external_stderr_url`"
            )

        return super().__new__(
            cls,
            log_key,
            external_stdout_url=external_stdout_url,
            external_stderr_url=external_stderr_url,
            external_url=external_url,
            shell_cmd=shell_cmd,
        )


class CapturedLogData(
    NamedTuple(
        "_CapturedLogData",
        [
            ("log_key", Sequence[str]),
            ("stdout", Optional[bytes]),
            ("stderr", Optional[bytes]),
            ("cursor", Optional[str]),
        ],
    )
):
    """Object representing captured log data, either a partial chunk of the log data or the full
    capture.  Contains the raw bytes and optionally the cursor offset for the partial chunk.
    """

    def __new__(
        cls,
        log_key: Sequence[str],
        stdout: Optional[bytes] = None,
        stderr: Optional[bytes] = None,
        cursor: Optional[str] = None,
    ):
        return super().__new__(cls, log_key, stdout, stderr, cursor)


class CapturedLogMetadata(
    NamedTuple(
        "_CapturedLogMetadata",
        [
            ("stdout_location", Optional[str]),
            ("stderr_location", Optional[str]),
            ("stdout_download_url", Optional[str]),
            ("stderr_download_url", Optional[str]),
        ],
    )
):
    """Object representing metadata info for the captured log data.
    It can contain:
     - a display string for the location of the log data,
     - a URL for direct download of the captured log data.
    """

    def __new__(
        cls,
        stdout_location: Optional[str] = None,
        stderr_location: Optional[str] = None,
        stdout_download_url: Optional[str] = None,
        stderr_download_url: Optional[str] = None,
    ):
        return super().__new__(
            cls,
            stdout_location=stdout_location,
            stderr_location=stderr_location,
            stdout_download_url=stdout_download_url,
            stderr_download_url=stderr_download_url,
        )


class CapturedLogSubscription:
    def __init__(
        self,
        manager: "ComputeLogManager[T_DagsterInstance]",
        log_key: Sequence[str],
        cursor: Optional[str],
    ):
        self._manager = manager
        self._log_key = log_key
        self._cursor = cursor
        self._observer: Optional[Callable[[CapturedLogData], None]] = None
        self.is_complete = False

    def __call__(self, observer: Optional[Callable[[CapturedLogData], None]]) -> Self:
        self._observer = observer
        self.fetch()
        if self._manager.is_capture_complete(self._log_key):
            self.complete()
        return self

    @property
    def log_key(self) -> Sequence[str]:
        return self._log_key

    def dispose(self) -> None:
        self._observer = None
        self._manager.unsubscribe(self)

    def fetch(self) -> None:
        if not self._observer:
            return

        should_fetch = True
        while should_fetch:
            log_data = self._manager.get_log_data(
                self._log_key,
                self._cursor,
                max_bytes=MAX_BYTES_CHUNK_READ,
            )
            if not self._cursor or log_data.cursor != self._cursor:
                self._observer(log_data)
                self._cursor = log_data.cursor
            should_fetch = _has_max_data(log_data.stdout) or _has_max_data(log_data.stderr)

    def complete(self) -> None:
        self.is_complete = True


def _has_max_data(chunk: Optional[bytes]) -> bool:
    # function is used as predicate but does not actually return a boolean
    return chunk and len(chunk) >= MAX_BYTES_CHUNK_READ  # type: ignore


class ComputeLogManager(ABC, MayHaveInstanceWeakref[T_DagsterInstance]):
    """Abstract base class for capturing the unstructured logs (stdout/stderr) in the current
    process, stored / retrieved with a provided log_key.
    """

    @abstractmethod
    @contextmanager
    def capture_logs(self, log_key: Sequence[str]) -> Generator[CapturedLogContext, None, None]:
        """Context manager for capturing the stdout/stderr within the current process, and persisting
        it under the given log key.

        Args:
            log_key (List[String]): The log key identifying the captured logs
        """

    @abstractmethod
    @contextmanager
    def open_log_stream(
        self, log_key: Sequence[str], io_type: ComputeIOType
    ) -> Iterator[Optional[IO[bytes]]]:
        """Context manager for providing an IO stream that enables the caller to write to a log stream
        managed by the captured log manager, to be read later using the given log key.

        Args:
            log_key (List[String]): The log key identifying the captured logs
        """

    @abstractmethod
    def is_capture_complete(self, log_key: Sequence[str]) -> bool:
        """Flag indicating when the log capture for a given log key has completed.

        Args:
            log_key (List[String]): The log key identifying the captured logs

        Returns:
            Boolean
        """

    @abstractmethod
    def get_log_data_for_type(
        self,
        log_key: Sequence[str],
        io_type: ComputeIOType,
        offset: int,
        max_bytes: Optional[int],
    ) -> tuple[Optional[bytes], int]:
        """Returns a chunk of the captured io_type logs for a given log key.

        Args:
            log_key (List[String]): The log key identifying the captured logs
            io_type (ComputeIOType): stderr or stdout
            offset (Optional[int]): An offset in to the log to start from
            max_bytes (Optional[int]): A limit on the size of the log chunk to fetch

        Returns:
            Tuple[Optional[bytes], int]: The content read and offset in to the file
        """

    @abstractmethod
    def get_log_metadata(self, log_key: Sequence[str]) -> CapturedLogMetadata:
        """Returns the metadata of the captured logs for a given log key, including
        displayable information on where the logs are persisted.

        Args:
            log_key (List[String]): The log key identifying the captured logs

        Returns:
            CapturedLogMetadata
        """

    @abstractmethod
    def delete_logs(
        self,
        log_key: Optional[Sequence[str]] = None,
        prefix: Optional[Sequence[str]] = None,
    ) -> None:
        """Deletes the captured logs for a given log key.

        Args:
            log_key(Optional[List[String]]): The log key of the logs to delete
            prefix(Optional[List[String]]): The prefix of the log keys to delete
        """

    @abstractmethod
    def subscribe(
        self, log_key: Sequence[str], cursor: Optional[str] = None
    ) -> CapturedLogSubscription:
        """Registers an observable object for log data.

        Args:
            log_key (List[String]): The log key identifying the captured logs
            cursor (Optional[String]): The string cursor marking the position within the log stream
        Returns:
            CapturedLogSubscription
        """

    def unsubscribe(self, subscription: CapturedLogSubscription) -> None:
        """Deregisters an observable object from receiving log updates.

        Args:
            subscription (CapturedLogSubscription): subscription object which manages when to send
                back data to the subscriber
        """
        pass

    def dispose(self):
        pass

    def parse_cursor(self, cursor: Optional[str] = None) -> tuple[int, int]:
        # Translates a string cursor into a set of byte offsets for stdout, stderr
        if not cursor:
            return 0, 0

        parts = cursor.split(":")
        if not parts or len(parts) != 2:
            return 0, 0

        stdout, stderr = [int(_) for _ in parts]
        return stdout, stderr

    def build_cursor(self, stdout_offset: int, stderr_offset: int) -> str:
        return f"{stdout_offset}:{stderr_offset}"

    def get_log_data(
        self,
        log_key: Sequence[str],
        cursor: Optional[str] = None,
        max_bytes: Optional[int] = None,
    ) -> CapturedLogData:
        stdout_offset, stderr_offset = self.parse_cursor(cursor)
        stdout, new_stdout_offset = self.get_log_data_for_type(
            log_key,
            ComputeIOType.STDOUT,
            stdout_offset,
            max_bytes,
        )
        stderr, new_stderr_offset = self.get_log_data_for_type(
            log_key,
            ComputeIOType.STDERR,
            stderr_offset,
            max_bytes,
        )
        return CapturedLogData(
            log_key=log_key,
            stdout=stdout,
            stderr=stderr,
            cursor=self.build_cursor(new_stdout_offset, new_stderr_offset),
        )

    def build_log_key_for_run(self, run_id: str, step_key: str) -> Sequence[str]:
        """Legacy adapter to translate run_id/key to captured log manager-based log_key."""
        return [run_id, "compute_logs", step_key]

    def get_log_keys_for_log_key_prefix(
        self, log_key_prefix: Sequence[str], io_type: ComputeIOType
    ) -> Sequence[Sequence[str]]:
        """Returns the logs keys for a given log key prefix. This is determined by looking at the
        directory defined by the log key prefix and creating a log_key for each file in the directory.
        """
        raise NotImplementedError("Must implement get_log_keys_for_log_key_prefix")

    def _get_log_lines_for_log_key(
        self,
        log_key: Sequence[str],
        io_type: ComputeIOType,
    ) -> Sequence[str]:
        """For a log key, gets the corresponding file, and splits the file into lines."""
        log_data, _ = self.get_log_data_for_type(
            log_key,
            io_type,
            offset=0,
            max_bytes=None,
        )
        raw_logs = log_data.decode("utf-8") if log_data else ""
        log_lines = raw_logs.split("\n")

        return log_lines

    def read_log_lines_for_log_key_prefix(
        self,
        log_key_prefix: Sequence[str],
        cursor: Optional[str],
        io_type: ComputeIOType,
    ) -> tuple[Sequence[str], Optional[LogLineCursor]]:
        """For a given directory defined by log_key_prefix that contains files, read the logs from the files
        as if they are a single continuous file. Reads env var DAGSTER_CAPTURED_LOG_CHUNK_SIZE lines at a time.
        Returns the lines read and the next cursor.

        Note that the has_more_now attribute of the cursor indicates if there are more logs that can be read immediately.
        If has_more_now if False, the process producing logs could still be running and dump more logs into the
        directory at a later time.
        """
        num_lines = int(os.getenv("DAGSTER_CAPTURED_LOG_CHUNK_SIZE", "1000"))
        # find all of the log_keys to read from and sort them in the order to be read
        log_keys = sorted(
            self.get_log_keys_for_log_key_prefix(log_key_prefix, io_type=io_type),
            key=lambda x: "/".join(x),
        )
        if len(log_keys) == 0:
            return [], None

        log_cursor = LogLineCursor.parse(cursor) if cursor else None
        if log_cursor is None:
            log_key_to_fetch_idx = 0
            line_cursor = 0
        else:
            log_key_to_fetch_idx = log_keys.index(log_cursor.log_key)
            line_cursor = log_cursor.line

        if line_cursor == -1:
            # line_cursor for -1 means the entirety of the file has been read, but the next file
            # didn't exist yet. So we see if a new file has been added.
            # if the next file doesn't exist yet, return
            if log_key_to_fetch_idx + 1 >= len(log_keys):
                return [], log_cursor
            log_key_to_fetch_idx += 1
            line_cursor = 0

        log_lines = self._get_log_lines_for_log_key(log_keys[log_key_to_fetch_idx], io_type=io_type)
        records = []
        has_more = True

        while len(records) < num_lines:
            remaining_log_lines = log_lines[line_cursor:]
            remaining_lines_to_fetch = num_lines - len(records)
            if remaining_lines_to_fetch < len(remaining_log_lines):
                records.extend(remaining_log_lines[:remaining_lines_to_fetch])
                line_cursor += remaining_lines_to_fetch
            else:
                records.extend(remaining_log_lines)
                line_cursor = -1

            if line_cursor == -1:
                # we've read the entirety of the file, update the cursor
                if log_key_to_fetch_idx + 1 >= len(log_keys):
                    # no more files to process
                    has_more = False
                    break
                log_key_to_fetch_idx += 1
                line_cursor = 0
                if len(records) < num_lines:
                    # we still need more records, so fetch the next file
                    log_lines = self._get_log_lines_for_log_key(
                        log_keys[log_key_to_fetch_idx], io_type=io_type
                    )

        new_cursor = LogLineCursor(
            log_key=log_keys[log_key_to_fetch_idx],
            line=line_cursor,
            has_more_now=has_more,
        )
        return records, new_cursor
